/*
 * Project: irond
 * Package: src.de.fhhannover.inform.iron.mapserver.communication.ifmap
 * File:    EventProcessor.java
 *
 * Copyright (C) 2010-2011 Fachhochschule Hannover
 * Ricklinger Stadtweg 118, 30459 Hannover, Germany 
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.fhhannover.inform.iron.mapserver.communication.ifmap;

import java.io.IOException;
import java.io.InputStream;

import org.apache.log4j.Logger;

import de.fhhannover.inform.iron.mapserver.communication.ChannelIdentifier;
import de.fhhannover.inform.iron.mapserver.communication.ClientIdentifier;
import de.fhhannover.inform.iron.mapserver.communication.bus.Processor;
import de.fhhannover.inform.iron.mapserver.communication.bus.Queue;
import de.fhhannover.inform.iron.mapserver.communication.bus.messages.Action;
import de.fhhannover.inform.iron.mapserver.communication.bus.messages.ActionSeries;
import de.fhhannover.inform.iron.mapserver.communication.bus.messages.BadChannelEvent;
import de.fhhannover.inform.iron.mapserver.communication.bus.messages.ClosedChannelEvent;
import de.fhhannover.inform.iron.mapserver.communication.bus.messages.Event;
import de.fhhannover.inform.iron.mapserver.communication.bus.messages.RequestChannelEvent;
import de.fhhannover.inform.iron.mapserver.communication.bus.messages.SendResponseAction;
import de.fhhannover.inform.iron.mapserver.communication.bus.messages.TimerExpiredEvent;
import de.fhhannover.inform.iron.mapserver.communication.http.ActionProcessor;
import de.fhhannover.inform.iron.mapserver.communication.ifmap.exceptions.AbortRequestException;
import de.fhhannover.inform.iron.mapserver.communication.ifmap.exceptions.AccessDeniedException;
import de.fhhannover.inform.iron.mapserver.datamodel.DataModelService;
import de.fhhannover.inform.iron.mapserver.datamodel.SubscriptionObserver;
import de.fhhannover.inform.iron.mapserver.datamodel.exceptions.IdentifierConstructionException;
import de.fhhannover.inform.iron.mapserver.datamodel.exceptions.InvalidFilterException;
import de.fhhannover.inform.iron.mapserver.datamodel.exceptions.InvalidMetadataException;
import de.fhhannover.inform.iron.mapserver.datamodel.exceptions.MetadataConstructionException;
import de.fhhannover.inform.iron.mapserver.datamodel.exceptions.NoPollResultAvailableException;
import de.fhhannover.inform.iron.mapserver.datamodel.exceptions.NoSuchPublisherException;
import de.fhhannover.inform.iron.mapserver.datamodel.exceptions.NoSuchSubscriptionException;
import de.fhhannover.inform.iron.mapserver.datamodel.exceptions.PublisherConstructionException;
import de.fhhannover.inform.iron.mapserver.datamodel.exceptions.PurgePublisherNoAllowedException;
import de.fhhannover.inform.iron.mapserver.datamodel.exceptions.RequestCreationException;
import de.fhhannover.inform.iron.mapserver.datamodel.exceptions.RunningSessionException;
import de.fhhannover.inform.iron.mapserver.logging.Logging;
import de.fhhannover.inform.iron.mapserver.messages.DumpRequest;
import de.fhhannover.inform.iron.mapserver.messages.DumpResult;
import de.fhhannover.inform.iron.mapserver.messages.EndSessionRequest;
import de.fhhannover.inform.iron.mapserver.messages.ErrorCode;
import de.fhhannover.inform.iron.mapserver.messages.NewSessionRequest;
import de.fhhannover.inform.iron.mapserver.messages.PollRequest;
import de.fhhannover.inform.iron.mapserver.messages.PollResult;
import de.fhhannover.inform.iron.mapserver.messages.PublishRequest;
import de.fhhannover.inform.iron.mapserver.messages.PurgePublisherRequest;
import de.fhhannover.inform.iron.mapserver.messages.RenewSessionRequest;
import de.fhhannover.inform.iron.mapserver.messages.Request;
import de.fhhannover.inform.iron.mapserver.messages.SearchRequest;
import de.fhhannover.inform.iron.mapserver.messages.SearchResult;
import de.fhhannover.inform.iron.mapserver.messages.SubscribeRequest;
import de.fhhannover.inform.iron.mapserver.provider.AuthorizationProvider;
import de.fhhannover.inform.iron.mapserver.provider.PublisherIdGenerator;
import de.fhhannover.inform.iron.mapserver.provider.PublisherIdProvider;
import de.fhhannover.inform.iron.mapserver.provider.ServerConfigurationProvider;
import de.fhhannover.inform.iron.mapserver.provider.SessionIdProvider;
import de.fhhannover.inform.iron.mapserver.provider.StorePublisherIdException;
import de.fhhannover.inform.iron.mapserver.utils.NullCheck;
import de.fhhannover.inform.iron.mapserver.utils.RequestTransformer;
import de.fhhannover.inform.iron.mapserver.utils.ResponseCreator;

/**
 * Process {@link Event} objects originating from the HTTP layer or SessionTimer.
 * 
 * <p>
 * Note:  Exceptions coming from the {@link SessionRepository} are sanity
 * checks which in the end indicate programming errors, don't be irritated by
 * all these try/catch clauses.
 * </p>
 * <p>
 * TODO: Somebody should really think about locking and stuff :-(
 * </p>
 * 
 * @author aw
 *
 */
public class EventProcessor extends Processor<Event> implements SubscriptionObserver {
	
	private static final String sProcName = "EventProcessor";
	
	private static Logger sLogger;
	static {
		sLogger = Logging.getTheLogger();
	}
	
	/**
	 * A queue where we can put in stuff
	 */
	private Queue<ActionSeries> mActionQueue;
	
	/**
	 * Provides us with {@link ClientIdentifier} publisher-id mappings.
	 */
	private PublisherIdProvider mPublisherIdProv;
	
	/**
	 * Provides us with "fresh" session-ids.
	 */
	private SessionIdProvider mSessionIdProv;
	
	/**
	 * Transforming incoming requests to {@link Request} objects.
	 */
	private RequestTransformer mRequestTransformer;

	/**
	 * Used to build the responses to be sent back.
	 */
	private ResponseCreator mResponseCreator;
	
	/**
	 * Provides configuration
	 */
	private ServerConfigurationProvider mServerConf;
	
	/**
	 * Stores Information about the IF-MAP sessions currently active
	 */
	private SessionRepository mSessionRep;
	
	/**
	 * Accessing the data model layer
	 */
	private DataModelService mDataModel;
	
	/**
	 * Used to create our the timers for the sessions.
	 */
	private SessionTimerFactory mSessionTimerFactory;
	
	/**
	 * Used to generate publisher-id's for new MAPCs
	 */
	private PublisherIdGenerator mPublisherIdGenerator;
	
	/**
	 * Used to check whether a MAPC is allowed to do changing operations
	 * like publish and purgePublisher.
	 */
	private AuthorizationProvider mAuthorizationProv;
	
	/**
	 * Used to protect the critical part, including {@link SessionRepository}
	 * and {@link DataModelService}
	 */
	private Object mBigProcessLock;

	
	/**
	 * This list is to be filled with all the results that need to be
	 * send out. It can be processed without holding the mBigProcessLock,
	 * so we get some more parallelism.
	 *
	private ThreadLocal<LinkedList<Result>> mLocalResultList = new ThreadLocal<LinkedList<Result>>() {
		
		@Override
		protected LinkedList<Result> initialValue() {
			return new LinkedList<Result>();
		}
		
	};
	*/
	

	public EventProcessor(Queue<Event> eventQueue, int workers, int forwarders) {
		super(eventQueue, workers, forwarders);
		mSessionRep = new SessionRepositoryImpl();
		mBigProcessLock = new Object();
	}
	

	/* (non-Javadoc)
	 * @see de.fhhannover.inform.iron.mapserver.communication.bus.Processor#start()
	 * 
	 * Overwrite to have some checks whether everything is here...
	 */
	@Override
	public void start() {
		NullCheck.check(mActionQueue, "action queue not initialized");
		NullCheck.check(mPublisherIdProv, "publisher-id provider not initialized");
		NullCheck.check(mSessionIdProv, "session-id provider not initialized");
		NullCheck.check(mRequestTransformer, "request transformator not initialized");
		NullCheck.check(mResponseCreator, "response creator not initialized");
		NullCheck.check(mServerConf, "server configuration not initialized");
		NullCheck.check(mSessionRep, "session repository not initialized");
		NullCheck.check(mDataModel, "data model not initialized");
		NullCheck.check(mSessionTimerFactory, "timer factory not initialized");
		NullCheck.check(mPublisherIdGenerator, "publisherid generator not initialized");
		NullCheck.check(mAuthorizationProv, "authorization provider not initialized");
		super.start();
		sLogger.info(sProcName + ": Running with " + getWorkersCount() + 
				" workers and " + getForwardersCount() + " forwarders");
	}
	
	/* (non-Javadoc)
	 * @see de.fhhannover.inform.iron.mapserver.communication.bus.Processor#stop()
	 * 
	 * Give us a logging statement.
	 */
	public void stop() {
		super.stop();
		sLogger.info(sProcName + " stopped");
	}

	/**
	 * Dispatches between {@link Event} implementations. This is the entry
	 * point for the {@link Event} processing.
	 */
	@Override
	public void processWork(Event e) {
		
		if (e instanceof RequestChannelEvent) {
			processRequestChannelEvent((RequestChannelEvent) e);
		} else if (e instanceof ClosedChannelEvent) {
			processClosedChannelEvent((ClosedChannelEvent) e);
		} else if (e instanceof TimerExpiredEvent) {
			processTimerExpiredEvent((TimerExpiredEvent) e);
		} else if (e instanceof BadChannelEvent) {
			processBadChannelEvent((BadChannelEvent) e);
		} else {
			sLogger.error("Unhandled Event type");
		}
	}

	/**
	 * Processes a {@link RequestChannelEvent}.
	 * 
	 * <ul>
	 * <li>Unmarshal the request</li>
	 * <li>Dispatch between SSRC and ARC</li>
	 * </ul>
	 * 
	 * <p>
	 * Note: We send out a InvalidIdentifier or InvalidMetadata error, before
	 * we even check if this guy has a {@link Session}. That is because we 
	 * detect these errors during unmarshalling and handle them before looking
	 * up the {@link Session}.
	 * </p>
	 * 
	 * @param event
	 */
	private void processRequestChannelEvent(RequestChannelEvent event) {
		ChannelIdentifier channelId = event.getChannelIdentifier();
		ClientIdentifier clientId = event.getClientIdentifier();
		Request request = null;
		
		sLogger.trace("EventProcessor: RequestChannelEvent on channel " + channelId);
		
		try {
			sLogger.trace(sProcName + ": Beginning transformation...");
			long start = System.currentTimeMillis();
			request = mRequestTransformer.transform(event.getRequestContent());
			long end= System.currentTimeMillis();
			sLogger.trace(sProcName + ": Transformation took " + (end - start) + " ms");
			
		// catch, handle, return, don't do anything special!
		} catch (RequestCreationException e) {
			// FIXME:
			// The problem is, that this can also be due to a "non soap message",
			// in this case we should maybe react differently?
			sLogger.error(sProcName + ": " + e.getMessage());
			respondWithError(channelId, ErrorCode.Failure, e.getMessage());
			return;
		} catch (IdentifierConstructionException e) {
			sLogger.warn(sProcName + ": Invalid identifier from client: " + e.getMessage());
			respondWithError(channelId, ErrorCode.InvalidIdentifier, e.getMessage());
			return;
		} catch (MetadataConstructionException e) {
			sLogger.warn(sProcName + ": Invalid metadata from client: " + e.getMessage());
			respondWithError(channelId, ErrorCode.InvalidMetadata, e.getMessage());
			return;
		} catch (InvalidFilterException e) {
			sLogger.warn(sProcName + ": Invalid filterstring from client: " + e.getMessage());
			respondWithError(channelId, ErrorCode.Failure, e.getMessage());
			return;
		}
		
		// At this point we start working with the repository and possibly with
		// the Datamodel, so this part is critical.
		
		// Only reached if unmarshalling was successful, dispatch between
		// SSRC and ARC
		synchronized (mBigProcessLock) {
			if (request instanceof PollRequest) {
				PollRequest pollReq = (PollRequest)request;
				processPollRequest(clientId, channelId, event.isFirstRequest(), pollReq);
			} else {
				processSsrcRequest(clientId, channelId, event.isFirstRequest(), request);
			}
		}
	}

	/**
	 * Dispatches between the different SSRC operations
	 * 
	 * @param clientId
	 * @param channelId
	 * @param request
	 */
	private void processSsrcRequest(ClientIdentifier clientId,
			ChannelIdentifier channelId, boolean firstRequest, Request request) {
		sLogger.trace(sProcName + ": Dispatching Operation for " + clientId + " on " + channelId);
		

		try {
			if (request instanceof NewSessionRequest) {
				sLogger.debug(sProcName + ": newSession for " + clientId
						+" on " + channelId);
				processNewSessionRequest(clientId, channelId, firstRequest,
						(NewSessionRequest) request);
			} else if (request instanceof EndSessionRequest) {
				sLogger.debug(sProcName + ": endSession for " + clientId
						+ " on " + channelId);
				processEndSessionRequest(clientId, channelId, firstRequest,
						(EndSessionRequest) request);
			} else if (request instanceof RenewSessionRequest) {
				sLogger.debug(sProcName + ": renewSession for " + clientId
						+ " on " + channelId);
				processRenewSessionRequest(clientId, channelId, firstRequest,
						(RenewSessionRequest) request);
			} else if (request instanceof PublishRequest) {
				sLogger.debug(sProcName + ": publish for " + clientId
						+ " on " + channelId);
				processPublishRequest(clientId, channelId, firstRequest,
						(PublishRequest) request);
			} else if (request instanceof SearchRequest) {
				sLogger.debug(sProcName + ": search for " + clientId
						+ " on " + channelId);
				processSearchRequest(clientId, channelId, firstRequest,
						(SearchRequest) request);
			} else if (request instanceof SubscribeRequest) {
				sLogger.debug(sProcName + ": subscribe for " + clientId
						+ " on " + channelId);
				processSubscribeRequest(clientId, channelId, firstRequest,
						(SubscribeRequest) request);
			} else if (request instanceof PurgePublisherRequest) {
				sLogger.debug(sProcName + ": purgePublisher for " + clientId
						+ " on " + channelId);
				processPurgePublisherRequest(clientId, channelId, firstRequest,
						(PurgePublisherRequest) request);
			} else if (request instanceof DumpRequest) {
				sLogger.debug(sProcName + ": dump for " + clientId
						+ " on " + channelId);
				processDumpRequest(clientId, channelId, firstRequest,
						(DumpRequest) request);
			} else {
				sLogger.error("Unhandled request: " + request.getClass().getName());
			}
		} catch (NoSuchPublisherException e) {
			// This exception comes from the DataModel and indicates that we
			// have a programming error somewhere, because we should never give
			// session-id's down to the DataModel, when there was no newSesson()
			// call for the DataModel. Respond with SystemError.
			sLogger.error("UNEXPECTED: DataModel doesn't know about the session" +
					" for " + clientId);
			respondWithError(channelId, ErrorCode.SystemError, "Internal State Failure");
		} catch (AccessDeniedException e) {
			sLogger.warn(sProcName + ": AccessDenied: " + e.getMessage() + " ("
					+ clientId + ")");
			respondWithError(channelId, ErrorCode.AccessDenied, e.getMessage());
		}
	}

	/**
	 * Processes a newSession request.
	 * 
	 * @param clientId
	 * @param channelId
	 * @param request
	 * @throws NoSuchPublisherException 
	 */
	private void processNewSessionRequest(ClientIdentifier clientId,
			ChannelIdentifier channelId, boolean firstRequest,
			NewSessionRequest request) throws NoSuchPublisherException {
		
		Session session = getSessionValidated(clientId);
		
		ActionSeries as = new ActionSeries();
	
		if (session != null) {
			// This guy has a running session, lets see...
			if (channelId.equals(session.getSsrc()) || session.getSsrc() == null || firstRequest) {
				sLogger.debug("newSession from " + clientId + " while having an" +
						" active session. That's fine, we end the old and create a new.");
				// the request is on the same channel or on a new one because the
				// old one was closed. Process it, but kill the old session before,
				// send endSessionResult on an existing ARC.
				if (session.isPollPending()) {
					sLogger.debug("Sending endSession on ARC to " + clientId);
					ChannelIdentifier arc = session.getArc();
					InputStream endSessionResp = mResponseCreator.createEndSessionResponse();
					as.add(new SendResponseAction(arc, endSessionResp));
				}
				
				// remove this session 
				endSessionLocal(session);
				session = null;
			} else {
				// Man, what are you trying to do?
				// Active session on channel1 and trying to open another one
				// on channel2... would you please stop doing that?
				// TODO: What to do? Close him away? Send him a reply?
				sLogger.warn(sProcName + ": " + clientId.toString() + " tries to" +
						" create a new session on a channel that was used before" +
						" We send AccessDenied.");

				// FIXME:
				// I don't know what to do in that case, because a client really
				// shouldn't do that. (It could happen because of a race condition, though)
				// We simply sent him an AccessDenied for now.
				respondWithError(channelId, ErrorCode.AccessDenied, "Only one SSRC allowed.");
				return;
			}
		}
		
		// If this guy had a session before, at this point this session is away.
		// Create the new one and register it.
		session = newSessionLocal(clientId, channelId, request);

		// construct the reply 
		String sessionId = session.getSessionId();
		String publisherId = session.getPublisherId();
		long reqMaxPollResultSize = request.getMaxPollResultSize();
		InputStream newSessionResp = null;
	
		// if the request contained a maxPollResultSize, return the one we selected
		if (reqMaxPollResultSize == NewSessionRequest.MAX_POLL_RESULT_NOT_GIVEN) {
			newSessionResp = mResponseCreator.createNewSessionResponse(sessionId,publisherId);
		} else {
			newSessionResp = mResponseCreator.createNewSessionResponse(sessionId,
					publisherId, session.getMaxPollResultSize());
		}
		as.add(new SendResponseAction(channelId, newSessionResp));
		
		// We should have a session now, send out the result... (including
		// a possible endSessionResponse on the old ARC)
		putActionSeriesIntoActionQueue(as);
	}

	/**
	 * Process an endSession request.
	 * 
	 * @param clientId
	 * @param channelId
	 * @param firstRequest 
	 * @param request
	 * @throws NoSuchPublisherException
	 */
	private void processEndSessionRequest(ClientIdentifier clientId,
			ChannelIdentifier channelId, boolean firstRequest, EndSessionRequest request) throws NoSuchPublisherException {
		try {
			ActionSeries as = new ActionSeries();
			
			Session session = getSessionValidated(clientId, channelId, request.getSessionId());
			
			checkAndSetSsrcChannelCorrect(channelId, firstRequest, session);
		
			// send endSessionResult on an existing ARC, if a poll is pending
			if (session.isPollPending()) {
				ChannelIdentifier arc = session.getArc();
				if (arc == null) {
					// this is a sanity check
					sLogger.error("UNEXPECTED: poll pending but no ARC associated");
				} else {
					InputStream endSessionResp = mResponseCreator.createEndSessionResponse();
					as.add(new SendResponseAction(arc, endSessionResp));
				}
			}
			
			endSessionLocal(session);
			
			// the endSessionResponse
			InputStream endIs = mResponseCreator.createEndSessionResponse();
			as.add(new SendResponseAction(channelId, endIs));
			
			// send out the endSessionResponse (including an optional one on the ARC)
			putActionSeriesIntoActionQueue(as);
			
		} catch (AbortRequestException e) {
			// abort
		}
	}

	/**
	 * Process a renewSession request.
	 * 
	 * @param clientId
	 * @param channelId
	 * @param firstRequest 
	 * @param request
	 * @throws NoSuchPublisherException
	 */
	private void processRenewSessionRequest(ClientIdentifier clientId,
			ChannelIdentifier channelId, boolean firstRequest,
			RenewSessionRequest request) throws NoSuchPublisherException {
		sLogger.debug("Processing renewSession");
		
			try {
				Session session = getSessionValidated(clientId, channelId,
						request.getSessionId());
				
				checkAndSetSsrcChannelCorrect(channelId, firstRequest, session);
				
				respondWithRenewSessionReceived(channelId);
			
			} catch (AbortRequestException e) {
				// abort
			}
	}

	/**
	 * Process a publish request.
	 * 
	 * @param clientId
	 * @param channelId
	 * @param request
	 * @throws NoSuchPublisherException
	 */
	private void processPublishRequest(ClientIdentifier clientId,
			ChannelIdentifier channelId, boolean firstRequest,
			PublishRequest request)
						throws NoSuchPublisherException, AccessDeniedException {
		
		try {
			if (!mAuthorizationProv.isWriteAllowed(clientId))
				throw new AccessDeniedException("read-only MAPC - dublish denied");
				
			Session session = getSessionValidated(clientId, channelId, request.getSessionId());
			checkAndSetSsrcChannelCorrect(channelId, firstRequest, session);
			
			mDataModel.publish(request);
			
			respondWithPublishReceived(channelId);
			
		} catch (AbortRequestException e) {
			// abort, error message is already on the way
		} catch (InvalidMetadataException e) {
			// this happens if that guy publishes contradictionary cardinality
			// for metadata, send him the error response.
			respondWithError(channelId, ErrorCode.InvalidMetadata, e.getMessage());
		}
	}
	
	/**
	 * Process a search request.
	 * 
	 * After we received the {@link SearchResult} from the {@link DataModelService},
	 * we marshal it. The max-size given in the {@link SearchRequest} is checked
	 * against the whole XML document to be sent out to the MAPC.
	 * 
	 * If a MAPC requested a max-size greater then the servers maximum supported
	 * size, we use the servers maximum.
	 * 
	 * @param clientId
	 * @param channelId
	 * @param firstRequest 
	 * @param request
	 * @throws NoSuchPublisherException 
	 */
	private void processSearchRequest(ClientIdentifier clientId,
			ChannelIdentifier channelId, boolean firstRequest, SearchRequest request) throws NoSuchPublisherException {
		try {
			Session session = getSessionValidated(clientId, channelId, request.getSessionId());
			checkAndSetSsrcChannelCorrect(channelId, firstRequest, session);
			
			SearchResult result = mDataModel.search(request);
	
			InputStream searchRes = mResponseCreator.createSearchResponse(result);
			
			int resultSize = 0;
			int servMaxSize = mServerConf.getDefaultMaxSearchResultSize();
			int maxSize = servMaxSize;
			int reqMaxSize = request.getMaxResultSize();
			
			try {
				resultSize = searchRes.available();
			} catch (IOException e) {
				sLogger.error("UNEXPECTED: Could not determine size of SearchResult");
			}
		
			// set maximum the one that may be stored, only if it is lower
			// than the servers max-size
			if (reqMaxSize != SearchRequest.MAX_RESULT_SIZE_NOT_GIVEN) {
				maxSize = (reqMaxSize < servMaxSize) ? reqMaxSize : servMaxSize;
			}
			
			if (resultSize > maxSize) {
				sLogger.warn(sProcName + ": SearchResultTooBig for " + clientId);
				respondWithError(channelId, ErrorCode.SearchResultsTooBig, 
						"max-size given was " + maxSize + " bytes, but result " +
								"size  was " + resultSize + " bytes.");
			} else {
				// size should be ok
				sLogger.trace("Putting searchResult into queue for " + clientId);
				putIntoQueue(new SendResponseAction(channelId, searchRes));
			}
		} catch (AbortRequestException e) {
			// abort
		}
	}
	
	/**
	 * Process a subscribe request.
	 * 
	 * @param clientId
	 * @param channelId
	 * @param request
	 * @throws NoSuchPublisherException
	 */
	private void processSubscribeRequest(ClientIdentifier clientId,
			ChannelIdentifier channelId, boolean firstRequest,
			SubscribeRequest request) throws NoSuchPublisherException {
		sLogger.trace("Processing SubscribeRequest for " + clientId);
		try {
			Session session = getSessionValidated(clientId, channelId, request.getSessionId());
			checkAndSetSsrcChannelCorrect(channelId, firstRequest, session);
			
			mDataModel.subscribe(request);
			
			respondWithSubscribeReceived(channelId);
		
		} catch (NoSuchSubscriptionException e) {
			// respond with some error?
			sLogger.warn(sProcName + ": " + e.getMessage() + " (" + clientId + ")");
			respondWithError(channelId, ErrorCode.Failure, e.getMessage());
		} catch (AbortRequestException e) {
			// abort
		}
	}
	
	/**
	 * Process a purgePublisher request.
	 * 
	 * @param clientId
	 * @param channelId
	 * @param request
	 * @throws NoSuchPublisherException
	 */
	private void processPurgePublisherRequest(ClientIdentifier clientId,
			ChannelIdentifier channelId, boolean firstRequest, PurgePublisherRequest request)
						throws NoSuchPublisherException, AccessDeniedException {
		try {
			if (!mAuthorizationProv.isWriteAllowed(clientId))
				throw new AccessDeniedException("read-only MAPC - purgePublisher denied");
			String sessionId = request.getSessionId();
			Session session = getSessionValidated(clientId, channelId, sessionId);
			checkAndSetSsrcChannelCorrect(channelId, firstRequest, session);
			
			mDataModel.purgePublisher(request.getSessionId(), request.getPublisherId());
			
			respondWithPurgePublisherReceived(channelId);

		} catch (AbortRequestException e) {
			// abort
		} catch (PurgePublisherNoAllowedException e) {
			sLogger.warn(sProcName + ": " + e.getMessage() + " (" + clientId + ")");
			respondWithError(channelId, ErrorCode.AccessDenied, e.getMessage());
		}
	}
	
	/**
	 * Process a dump request.
	 * 
	 * @param clientId
	 * @param channelId
	 * @param request
	 */
	private void processDumpRequest(ClientIdentifier clientId,
			ChannelIdentifier channelId, boolean firstRequest, DumpRequest request) {				
		try {
			Session session = getSessionValidated(clientId, channelId, request.getSessionId());
			checkAndSetSsrcChannelCorrect(channelId, firstRequest, session);
			DumpResult result = mDataModel.dump(session.getSessionId());
			result.setFilter(request.getIdentifier());
			InputStream dumpRes = mResponseCreator.createDumpResponse(result);
			putIntoQueue(new SendResponseAction(channelId, dumpRes));
		} catch (AbortRequestException e) {
			sLogger.info("Dump request aborted.");
		}		
	}
	
	/**
	 * Process a poll request.
	 * 
	 * We try to find the {@link Session} for the client doing the poll request.
	 * If there is a {@link PollResult} available, put it into the {@link Queue}
	 * and reset the state. Else, we mark a pending poll for the {@link Session}.
	 * 
	 * @see EventProcessor#getSessionValidated(ClientIdentifier)
	 * @see EventProcessor#handleDoubleArcFailure(ChannelIdentifier, Session)
	 * @see EventProcessor#handleSendPollRes(ChannelIdentifier, Session)
	 * @see EventProcessor#handlePollPending(ChannelIdentifier, Session)
	 * 
	 * @param request
	 * @param arc
	 * @param clientId
	 */
	private void processPollRequest(ClientIdentifier clientId,
			ChannelIdentifier arc, boolean firstRequest, PollRequest request) {

		String sessionId = request.getSessionId();
		
		try {
			Session session = getSessionValidated(clientId, arc, sessionId);
			sLogger.trace(sProcName + ": PollRequest from " + clientId + " on " +
					arc);
			
			checkAndSetArc(arc, firstRequest, session);
			
			if (session.isPollResAvail()) {
				handleSendPollRes(arc, session);
			} else {
				handlePollPending(arc, session);
			}
		} catch (AbortRequestException e) { 
			// the client is already notified
		} catch (NoSuchPublisherException e) {
			// this occurs if the data model does not know about the
			// given session-id. This should not happen.
			sLogger.error("UNEXPECTED: DataModel does not know about the session.");
			respondWithError(arc, ErrorCode.SystemError, "Internal State Failure");
		}
	}
	
	/**
	 * This method takes care of the case when a client has opened a second
	 * ARC while a poll is still pending.
	 * 
	 * The following steps are done:
	 * 
	 * <ul>
	 * <li>end the session session</li>
	 * <li>respond to the poll on the older ARC with end session result</li>
	 * <li>respond to the poll on the new ARC with errorResult InvalidSessionId</li>
	 * </ul>
	 * 
	 * Don't send anything out on the SSRC, just remove the session, we are not
	 * able to do so, anyway.
	 * 
	 * TODO: Should we instruct the HTTP-layer to remove the channel?
	 *       Better not, because the client may want to reuse it?
	 * 
	 * @param newArc
	 * @param session
	 * @throws NoSuchPublisherException 
	 * @throws AbortRequestException 
	 */
	private void handleDoubleArcFailure(ChannelIdentifier newArc, Session session) throws NoSuchPublisherException, AbortRequestException {
		ChannelIdentifier oldArc = session.getArc();
		sLogger.warn(sProcName + ": " + session.getClientIdentifier() + 
				"created a second ARC while another poll was pending. Ending the " +
				"session and sending responses out.");
		
		endSessionLocal(session);
		
		ActionSeries as = new ActionSeries();
		InputStream endSessionResult = mResponseCreator.createEndSessionResponse();
		InputStream errorResponse = mResponseCreator.createErrorResponse(
				ErrorCode.InvalidSessionId, "two times ARC");
		as.add(new SendResponseAction(oldArc, endSessionResult));
		as.add(new SendResponseAction(newArc, errorResponse));
		putActionSeriesIntoActionQueue(as);
	}
	
	/**
	 * Sends a {@link PollResult} to a client and clears the state of the
	 * {@link Session} regarding pending polls afterwards.
	 * If sending is not possible because the {@link PollResult} is too big,
	 * we call call a helper.
	 * 
	 * @see EventProcessor#handlePollResultTooBig(ChannelIdentifier, Session);
	 * 
	 * @param arc the channel where the {@link PollResult} ist to be send out
	 * @param session the corresponding session.
	 */
	private void handleSendPollRes(ChannelIdentifier arc, Session session) {
		try {
			String sessionId = session.getSessionId();
			long pollSize = session.getMaxPollResultSize();
			
			// sanity check
			if (session.hasArc() && !session.getArc().equals(arc)) {
				sLogger.error("UNEXPECTED: Missed a double ARC failure...");
				// TODO: Do some cleanup?
			}
			
			// get reply and put it into the local list.
			PollResult pollResult = mDataModel.getPollResultFor(sessionId);
			InputStream pollIs = mResponseCreator.createPollResponse(pollResult);
			clearPollState(session);
			
			// check if the PollResult is too big to be sent out.
			try {
				pollSize = pollIs.available();
			} catch (IOException e) {
				// If this fails nothing is going to work in turn, anyway.
				// pollIs is of type ByteArrayInputStream, so we should be able
				// to get the number of available bytes without getting an
				// IOException, but you never know.
				sLogger.error("UNEXPECTED: Not able to determine size of PollResult"); 
			}
		
			if (pollSize > session.getMaxPollResultSize()) {
				sLogger.debug("PollResult too big for " + session.getClientIdentifier());
				handlePollResultTooBig(arc, session);
				return;
			}
			
			// send out the reply (asynchronous)
			sLogger.trace("Sending out PollResult to " + session.getClientIdentifier());
			putIntoQueue(new SendResponseAction(arc, pollIs));

		} catch (NoPollResultAvailableException e) {
			// Most likely this is a programming error, try to do some workaround:
			// Don't send anything back and set poll pending state for this
			// session.
			sLogger.error("UNEXEPECTED: poll avl set in session, but no PollResult" +
					" found in  DataModel");
			handlePollPending(arc, session);
		} catch (NoSuchPublisherException e) {
			// this occurs if the data model does not know about the
			// given session-id. This should not happen.
			sLogger.error("UNEXPECTED: DataModel does not know about the session.");
			respondWithError(arc, ErrorCode.SystemError, "Internal State Failure");
		}
	}

	/**
	 * Handles the situation if a {@link PollResult} is too big.
	 * 
	 * If a the poll result was too big, we send out an error on the arc
	 * and reset the poll state of the session. Further, we have to inform
	 * the {@link DataModelService} to drop all active subscriptions for
	 * this client.
	 * 
	 * @param arc
	 * @param session
	 */
	private void handlePollResultTooBig(ChannelIdentifier arc, Session session) {
		try {
			mDataModel.removeSubscriptions(session.getSessionId());
		} catch (NoSuchPublisherException e) {
			sLogger.error("UNEXPECTED: DataModel doesn't know about the session" +
					" for " + session.getClientIdentifier());
			respondWithError(arc, ErrorCode.SystemError, "Internal State Failure");
		}
		clearPollState(session);
		sLogger.debug("Sending PollResultsTooBig to " + session.getClientIdentifier());
		respondWithPollResultError(arc, ErrorCode.PollResultsTooBig, "PollResults were too big");
	}

	/**
	 * Sets the state of the {@link Session} such that a pending poll
	 * is indicated. Map the ARC on which the poll is pending to the
	 * {@link Session} using the {@link SessionRepository}.
	 * If we had a timer running because the SSRC is closed, cancel this timer.
	 * 
	 * @param arc
	 * @param session
	 */
	private void handlePollPending(ChannelIdentifier arc, Session session) {
		NullCheck.check(session, "session is null");
		NullCheck.check(arc, "arc is null");
		sLogger.trace(sProcName + ": Setting poll pending for " + session.getClientIdentifier()
		+ " on " + arc);
		session.setPollPending(true);
		session.setArc(arc);
	}	

	/**
	 * Process a {@link ClosedChannelEvent}.
	 * 
	 * TODO: Some documentation.
	 * 
	 * @param e
	 */
	private void processClosedChannelEvent(ClosedChannelEvent e) {
		synchronized (mBigProcessLock) {
			ChannelIdentifier channel = e.getChannelIdentifier();
			sLogger.debug(sProcName + ": Got ClosedChannelEvent for " + channel);
			
			// try to look up the session associated with this channel.
			
			Session session = mSessionRep.getBy(channel);
		
			// if we don't find an associated session, we don't care
			if (session == null) {
				sLogger.trace("No session found for channel");
				return;
			}
			ChannelIdentifier arc = session.getArc();
			ChannelIdentifier ssrc = session.getSsrc();
			
		
			if (channel.equals(arc) && session.isPollPending()) {
				// in this case we have a problem, because the ARC was closed
				// while a poll was pending
				sLogger.warn(sProcName + ": ARC for " + session.getClientIdentifier() + 
						" closed while poll was pending. Calling endSession.");
				try {
					endSessionLocal(session);
				} catch (NoSuchPublisherException e1) {
					sLogger.error("UNEXPECTED: DataModel doesn't know about the session" +
						" for " + session.getClientIdentifier());
				}
				// No need to start a timer, the Session is gone anyway
				return;
				
			} else if (channel.equals(arc) && !session.isPollPending()) {
				// the ARC was closed, but that's ok, no poll was pending
				session.setArc(null);
				mSessionRep.unmap(session, channel);
			} else if (channel.equals(ssrc)) {
				// the SSRC was closed
				session.setSsrc(null);
				mSessionRep.unmap(session, channel);
			}
			
			if (session.hasTimer()) {
				sLogger.error("UNEXPECTED: Have a running timer for a Session" +
						" which just had a channel associated");
				session.getTimer().cancel();
				session.setTimer(null);
			}
			
			// only if we have no channel associated, start a timer.
			
			arc = session.getArc();
			ssrc = session.getSsrc();
			
			if (arc == null && ssrc == null) {
				sLogger.debug("Creating Timer for " + session.getClientIdentifier());
				String sessionId = session.getSessionId();
				session.setTimer(mSessionTimerFactory.newTimer(sessionId));
				session.getTimer().start();
			}
		}
	}

	/**
	 * Process a {@link TimerExpiredEvent}.
	 * 
	 * We try to find the corresponding {@link Session} with the given
	 * session-id and then run a normal endSession.
	 * 
	 * @param e
	 */
	private void processTimerExpiredEvent(TimerExpiredEvent e) {
		synchronized (mBigProcessLock) {
			sLogger.debug("Processing TimerEvent with session-id " + e.getSessionId());
			
			Session session = mSessionRep.getBy(e.getSessionId());
			
			if (session == null) {
				// That's weird
				sLogger.warn(sProcName + ": Could not find a associated session with" +
						" the TimerEvent");
				return;
			}
			
			// Do checks to see if the timer is really expired, or if it's just
			// a nasty race condition.
			if (!session.hasTimer()) {
				if (!session.hasSsrc() && !session.hasArc())
					sLogger.error("UNEXPECTED: Timer should be running for this session");
				
				return;
				
			} else if (session.hasTimer() && session.getTimer().isRunning()) {
				// huh... there is a timer, but it is still running. Can happen due
				// to racing of events, shouldn't happen too often, though.
				return;
			}
		
			sLogger.debug("Session for " + session.getClientIdentifier() + 
					" will be ended because of a timeout.");
			// Send endSessionResult on an existing ARC, if a poll is pending
			// This  should not happen, becase then a Timer wasn't created in the
			// first place, anyway
			if (session.isPollPending()) {
				sLogger.error("UNEXPECTED: Timeout for a session having a poll pending");
				ChannelIdentifier arc = session.getArc();
				InputStream endSessionResp = mResponseCreator.createEndSessionResponse();
				putIntoQueue(new SendResponseAction(arc, endSessionResp));
			}
			
			try {
				endSessionLocal(session);
			} catch (NoSuchPublisherException e1) {
				sLogger.error("UNEXPECTED: DataModel doesn't know about the session" +
						" for " + session.getClientIdentifier());
			}
		}
	}
	
	/**
	 * Process a {@link BadChannelEvent} e
	 * 
	 * This event can happen if a MAPC tries to do a request while the channel
	 * expected a response, something fails with a request on the channel or
	 * sending of a response on a channel fails.
	 * 
	 * We handle all of these cases by ending the session, optionally sending
	 * a endSessionResult on an ARC with a pending poll if the ARC is not the
	 * channel that caused the problem.
	 * 
	 * The channel that caused the problem is going to be closed by the
	 * http layer. If a {@link Session} had two channels attached, the second
	 * channel is not explicitly closed. Therefore an unaffected  SSRC or ARC
	 * may later be reused, given a new session is created by the MPAC.
	 * 
	 * @param e
	 */
	private void processBadChannelEvent(BadChannelEvent e) {
		synchronized (mBigProcessLock) {
			ChannelIdentifier chId = e.getChannelIdentifier();
			sLogger.debug(sProcName + ": Processing BadChannelEvent on " + chId);
		
			Session session = mSessionRep.getBy(chId);
		
			// do nothing if no session is known for this channel
			if (session == null) {
				sLogger.debug(sProcName + ": No session found for " + chId);
				return;
			}
			
			sLogger.trace(sProcName + ": Session for " + session.getClientIdentifier() +
					" will be closed because of a BadChannelEvent");
		
		
			// if this session has an ARC with a poll pending, and the channel which
			// caused the problem is not the ARC, send out an endSessionResult.
			if (session.hasArc() &&  !session.getArc().equals(chId) && session.isPollPending()) {
				InputStream endSessionResp = mResponseCreator.createEndSessionResponse();
				putIntoQueue(new SendResponseAction(session.getArc(), endSessionResp));
			}
		
			// forget the session
			try {
				endSessionLocal(session);
			} catch (NoSuchPublisherException e1) {
				sLogger.error(sProcName + ": UNEXPECTED: DataModel doesn't know " +
						"about the session for " + session.getClientIdentifier());
			}
		}
	}
	
	/**
	 * Get the right {@link Session} object from the {@link SessionRepository}.
	 * 
	 * We identify a MAPC by the session-id and the {@link ClientIdentifier}
	 * we get.
	 * If no session is stored belonging to the {@link ClientIdentifier} we
	 * send a AccessDenied error.
	 * If the MAPC uses a session-id belonging to a different MAPC we send
	 * a AccessDenied error.
	 * If we don't find a {@link Session} object for either the session-id or
	 * the {@link ClientIdentifier} we send InvalidSessionId.
	 * If we find a {@link Session} for the {@link ClientIdentifier} but not
	 * for the session-id given we send a InvalidSessionId, but indicate in
	 * the error string that a {@link Session} for this client exists.
	 * 
	 * If this method throws a {@link AbortRequestException} an error response
	 * was created and is on the way to the client.
	 * <b>Dont't put another {@link ActionSeries} into the action queue!</b>
	 * the request, don't send anything back!
	 * 
	 * @param clientId {@link ClientIdentifier} of the {@link Event}
	 * @param channelId {@link ChannelIdentifier} of the {@link Event}
	 * @param sessionId session-id found in the {@link Request} we are processing
	 * @return the {@link Session} for this MAPC or null in case no {@link Session}
	 *         was found.
	 * @throws AbortRequestException if the MAPC did something wrong.
	 */
	private Session getSessionValidated(ClientIdentifier clientId,
			ChannelIdentifier channelId, String sessionId) throws AbortRequestException {
		
		Session sessionClientId = mSessionRep.getBy(clientId);
		
		// special case for the other getSessionValidated() method
		if (sessionId == null && channelId == null)
			return sessionClientId;
		
		Session sessionSessionId = mSessionRep.getBy(sessionId);
	
		if (sessionClientId != sessionSessionId) {
			// this shouldn't be the case 
			if (sessionSessionId != null) {
				// Now that's fishy, this guy knows about a session-id that's
				// from a different client...
				sLogger.warn(sProcName + ": Ughh... Why does \"" + clientId
						+ "\" know about the session  of \""  
						+ sessionSessionId.getClientIdentifier() + "\" ?");
	
				// Let us be friendly and tell him that he just found a valid
				// session-id. We shouldn't do that, should we?
				respondWithError(channelId, ErrorCode.InvalidSessionId,
						"You are _not_ supposed to use the session-id of " +
						" somebody else.");
			} else {
				// sessionSessionId is null, so sessionClientId can't be.
				// Therefore we found a session for the MAPC, but the wrong
				// session-id was used. Transpose digits?
				sLogger.warn(sProcName + ": Client " + clientId + " uses wrong session-id");
				respondWithError(channelId, ErrorCode.InvalidSessionId,
						"You have a session, but used the wrong session-id.");
			}
		} else {
			// That's fine both are the same
			if (sessionSessionId == null) {
				sLogger.warn(sProcName + ": " + clientId + " tried to operate "
						+ "without having an active session");
				// Now that's not so cool, both are null, indicating neither
				// for the session-id nor for the clientId is a Session in the
				// repository. Send InvalidSessionId.
				respondWithError(channelId, ErrorCode.InvalidSessionId,
						"Session not found.");
			} else {
				// We found a valid Session, return it;
				return sessionSessionId;
			}
		}
		
		// Ending up here means we did a responseWithError() call, throw the
		// AbortRequestException.
		throw new AbortRequestException();
	}

	private Session getSessionValidated(ClientIdentifier clientId) {
		try {
			return getSessionValidated(clientId, null, null);
		} catch (AbortRequestException e) {
			// this should not be reached
			throw new RuntimeException("Programming error");
		}
	}
	
	/**
	 * Creates a new {@link Session}, stores it, creates mappings and
	 * returns the new created {@link Session}.
	 * If now max-poll-result-size was given, get the one from the 
	 * server configuration.
	 * If the {@link PublisherIdProvider} does not know about the {@link ClientIdentifier}
	 * we generate a new publisher-id and store it.
	 * 
	 * @param clientId
	 * @param channelId
	 * @param request
	 * @return
	 */
	private Session newSessionLocal(ClientIdentifier clientId,
			ChannelIdentifier channelId, NewSessionRequest request) {
		String publisherId = mPublisherIdProv.getPublisherIdFor(clientId);
	
		// if no publisherId is available we have to ask the generator to create
		// a new one.
		if (publisherId == null) {
			publisherId = mPublisherIdGenerator.generatePublisherIdFor(clientId,
					mPublisherIdProv);
			NullCheck.check(publisherId, "generated publisher-id is null");
			
			try {
				mPublisherIdProv.storePublisherIdFor(clientId, publisherId);
			} catch (StorePublisherIdException e) {
				sLogger.fatal("Could not store generated publisher-id for " + clientId);
				// we continue anyway
			}
		}
		
		String sessionId = mSessionIdProv.getSessionId();
		Session session = new Session(clientId, publisherId);
		
		if (request.getMaxPollResultSize() == NewSessionRequest.MAX_POLL_RESULT_NOT_GIVEN) {
			session.setMaxPollResultSize(mServerConf.getDefaultMaxPollResultSize());
		} else {
			// TODO: Check against a maximum? and send a error back in case
			// something goes wrong?
			session.setMaxPollResultSize(request.getMaxPollResultSize());
		}
		
		session.setSsrc(channelId);
		session.setSessionId(sessionId);
		
		mSessionRep.store(session);
		mSessionRep.map(session, channelId);
		mSessionRep.map(session, sessionId);
		
		// tell the data model
		try {
			mDataModel.newSession(sessionId, publisherId);
		} catch (RunningSessionException e) {
			sLogger.error("UNEXPECTED: Datamodel thinks there exists a "
					+ "session for this publisherId");
			sLogger.error(e.getMessage());
		} catch (PublisherConstructionException e) {
			sLogger.error("UNEXPECTED: Datamodel was unable to construct" +
					" a publisher Object");
			sLogger.error(e.getMessage());
		}
		
		return session;
	}
	
	/**
	 * Cleans up traces of a {@link Session}
	 * 
	 * @param session the {@link Session} to forget
	 * @throws NoSuchPublisherException 
	 */
	private void endSessionLocal(Session session) throws NoSuchPublisherException {
		
		if (session.hasArc()) mSessionRep.unmap(session, session.getArc());
		
		if (session.hasSsrc()) mSessionRep.unmap(session, session.getSsrc());
		
		if (session.hasTimer()) session.getTimer().cancel();
		
		String sessionId = session.getSessionId();
		if (sessionId != null) {
			mSessionRep.unmap(session, sessionId);
		// forward call to data model layer
			mDataModel.endSession(session.getSessionId());
		} else {
			sLogger.error("UNEXPECTED: Session without session-id for "
					+ session.getClientIdentifier() + " found");
		}
	
		// forget about it
		mSessionRep.drop(session);
	}
	
	/**
	 * Update the SSRC if the MAPC is starting to uese a new connection
	 * for the SSRC.
	 * 
	 * We also do canceling of the {@link SessionTimer} here, so this
	 * method can be used to implement renewSession.
	 *  
	 * @param channelId
	 * @param session
	 * @param request
	 * @throws AbortRequestException 
	 */
	private void checkAndSetSsrcChannelCorrect(ChannelIdentifier channelId,
			boolean firstRequest, Session session) throws AbortRequestException {
		NullCheck.check(session, "session is null");
		
		SessionTimer timer = session.getTimer();
		
		checkSsrcNotArc(channelId, session);
		
		if (firstRequest && !channelId.equals(session.getSsrc())) {
			sLogger.trace(sProcName + ": " + session.getClientIdentifier()
					+ " uses " + channelId + " as new SSRC");
			
			if (session.hasSsrc()) {
				mSessionRep.unmap(session, session.getSsrc());
				session.setSsrc(null);
			}

			// map in the new SSRC
			session.setSsrc(channelId);
			mSessionRep.map(session, channelId);
			
			if (session.hasTimer()) {
				sLogger.trace("Cancel timer for " + session.getClientIdentifier());
				timer.cancel();
				session.setTimer(null);
			}
			
		} else if (channelId.equals(session.getSsrc())) {
			sLogger.trace(sProcName + ": " + session.getClientIdentifier()
					+ " uses old SSRC for request");
			// that's good, we don't have to do anything
		} else {
			sLogger.warn(sProcName + ": " + session.getClientIdentifier() +
					" uses an old channel as SSRC. Sending out AccessDenied");
			respondWithError(channelId, ErrorCode.AccessDenied, "Existing SSRC");
			throw new AbortRequestException();
			
			
		}
	}
	
	/**
	 * Check if a client uses a good ARC to send us a poll request.
	 * If the SSRC is used send an error.
	 * If a second ARC is used while a poll is pending,
	 * {@link #handleDoubleArcFailure(ChannelIdentifier, Session)}
	 * manages everything.
	 * If no poll is pending, but a second ARC is used we send a AccessDenied
	 * "Existing ARC" error.
	 * 
	 * @param arc
	 * @param session
	 * @throws AbortRequestException 
	 * @throws NoSuchPublisherException 
	 */
	private void checkAndSetArc(ChannelIdentifier arc, boolean firstRequest,
			Session session) throws AbortRequestException, NoSuchPublisherException {
		
		checkArcNotSsrc(arc, session);
	
		if (session.hasArc() && session.isPollPending()) {
				handleDoubleArcFailure(arc, session);
				throw new AbortRequestException();
		} else if (firstRequest && !arc.equals(session.getArc())) {
			sLogger.trace(sProcName + ": " + session.getClientIdentifier()
					+ "uses " + arc + " as new ARC");
		
			// unmap the old ARC
			if (session.hasArc())
				mSessionRep.unmap(session, session.getArc());

			session.setArc(arc);
			mSessionRep.map(session, arc);
			
			if (session.hasTimer()) {
				session.getTimer().cancel();
				session.setTimer(null);
			}
		} else if (arc.equals(session.getArc())) {
			sLogger.trace(sProcName + ": " + session.getClientIdentifier()
					+ " uses old ARC for request");
			// that's good, we don't have to do anything
		} else {
			sLogger.warn(sProcName + ": " + session.getClientIdentifier() +
					" uses an old channel as ARC. Sending out AccessDenied");
			respondWithError(arc, ErrorCode.AccessDenied, "Existing SSRC");
			throw new AbortRequestException();
		}
	}
	
	/**
	 * Check if a client used its SSRC to send us a Poll request. If so,
	 * send an Error / Failure.
	 * 
	 * @param arc
	 * @param session
	 * @throws AbortRequestException
	 */
	private void checkArcNotSsrc(ChannelIdentifier arc, Session session) throws AbortRequestException {
		if (session.hasSsrc()) {
			if (session.getSsrc().equals(arc)) {
				sLogger.warn(sProcName + ": " + session.getClientIdentifier() +
						" uses its SSRC as ARC. Sending out a errorResult/Failure");
				respondWithError(arc, ErrorCode.Failure, "SSRC used as ARC");
				throw new AbortRequestException();
			}
		}
	}


	/**
	 * Check if a client used its ARC to send us a SSRC operation.
	 * If so, send an Error / Failure.
	 * 
	 * @param arc
	 * @param session
	 * @throws AbortRequestException 
	 */
	private void checkSsrcNotArc(ChannelIdentifier ssrc, Session session) throws AbortRequestException {
		if (session.hasArc()) {
			if (session.getArc().equals(ssrc)) {
				sLogger.warn(sProcName + ": " + session.getClientIdentifier() + 
						" uses its ARC as SSRC. Sending out a errorResult/Failure");
				respondWithError(session.getArc(), ErrorCode.Failure, "ARC used as SSRC");
				throw new AbortRequestException();
			}
		}
	}
	
	/* (non-Javadoc)
	 * @see de.fhhannover.inform.iron.mapserver.datamodel.SubscriptionObserver#pollResultAvailable(java.lang.String)
	 * 
	 * This method is called if the Datamodel wants to inform us about a available
	 * PollResult. We either send a PollResult out, if the session has a poll
	 * pending, or we mark the session has having a poll result available.
	 */
	@Override
	public void pollResultAvailable(String sessionId) {
		sLogger.trace("Datamodel callback pollResultAvailable: " + sessionId);
		
		Session session = mSessionRep.getBy(sessionId);
		
		if (session == null) {
			sLogger.error("UNEXPECTED: Datamodel has a session we don't know about");
			// We can't do much about it, it is simply a programming error
			// somewhere
			// We could end the session, but doing so during the callback is
			// not really a good thing, in fact I don't really know what will
			// happen.
		} else {
			if (session.isPollPending()) {
				// is waiting for a response of a previous poll request which  is
				// still pending
				ChannelIdentifier arc = session.getArc();
				if (arc == null) {
					// this should be never the case if isPollPending() is true.
					sLogger.error("UNEXPECTED: We have a pending poll but no" +
							" ARC associated");
					// can we do something about it?
					// for now, reset the poll state and mark it having a
					// poll result available. Resetting.
					sLogger.warn(sProcName + ": Clearing PollState and setting" +
							" PollResAvail for client " + session.getClientIdentifier());
					clearPollState(session);
					setPollStateResAvail(session);
				} else {
					// send the poll result out directly
					handleSendPollRes(arc, session);
				}
			} else {
				setPollStateResAvail(session);
			}
		}
	}
	
	
	
	
	/**********************************************************************
	 * From here on everything is trivial                                 *
	 **********************************************************************/
	
	/**
	 * Simple helper to reset the poll state of a {@link Session}, further
	 * we unmap the ARC from the session repository.
	 * 
	 * @param session
	 */
	private void clearPollState(Session session) {
		NullCheck.check(session, "session is null");
		session.setPollPending(false);
		session.setPollResAvail(false);
	}
	
	/**
	 * Simple helper to set the state of a {@link Session} such that
	 * it indicates an available poll result.
	 * 
	 * @param session
	 */
	private void setPollStateResAvail(Session session) {
		NullCheck.check(session, "session is null");
		session.setPollResAvail(true);
		
		if (session.isPollPending()) {
			sLogger.error("UNEXPECTED: PollPending and PollResAvailable is to be set.");
		}
	}

	/**
	 * Send an error response on the given channel.
	 * 
	 * This is a asynchronous call. There is by no means any guarantee
	 * that the message was delivered successfully to the client if this
	 * method returns. It's more like fire and forget.
	 * 
	 * @param channelId
	 * @param errCode
	 * @param errMsg
	 */
	private void respondWithError(ChannelIdentifier channelId, ErrorCode errCode,
			String errMsg) {
		InputStream resp = mResponseCreator.createErrorResponse(errCode, errMsg);
		putIntoQueue(new SendResponseAction(channelId, resp));
	}
	
	/**
	 * Send out an error response contained in a PollResult on the given Channel,
	 * possibly with a name for the subscription
	 * 
	 * @param channelId
	 * @param errCode
	 * @param errMsg
	 * @param subName
	 */
	private void respondWithPollResultError(ChannelIdentifier channelId,
			ErrorCode errCode, String errMsg, String subName) {
		InputStream resp = mResponseCreator.createPollResponseError(errCode,
				errMsg, subName);
		putIntoQueue(new SendResponseAction(channelId, resp));
	}

	/**
	 * Send out an error response contained in a PollResult on the given Channel,
	 * without a subscription name.
	 * 
	 * @param channelId
	 * @param errCode
	 * @param errMsg
	 */
	private void respondWithPollResultError(ChannelIdentifier channelId,
			ErrorCode errCode, String errMsg) {
		respondWithPollResultError(channelId, errCode, errMsg, null);
	}
	
	/**
	 * Put a single {@link Action} into the queue to send a publishReceived
	 * response on the channel identified by channelId.
	 * 
	 * @param channelId
	 */
	private void respondWithPublishReceived(ChannelIdentifier channelId) {
			InputStream pubRecv = mResponseCreator.createPublishResponse();
			putIntoQueue(new SendResponseAction(channelId, pubRecv));
	}
	
	/**
	 * Put a single {@link Action} into the queue to send a subscribeReceived
	 * response on the channel identified by channelId.
	 * 
	 * @param channelId
	 */
	private void respondWithSubscribeReceived(ChannelIdentifier channelId) {
		InputStream subscribeRes = mResponseCreator.createSubscribeResponse();
		putIntoQueue(new SendResponseAction(channelId, subscribeRes));
	}
	
	/**
	 * Put a single {@link Action} into the queue to send a purgePublisherReceived
	 * response on the channel identified by channelId.
	 * 
	 * @param channelId
	 */
	private void respondWithPurgePublisherReceived(ChannelIdentifier channelId) {
		InputStream reply = mResponseCreator.createPurgePublisherResponse();
		putIntoQueue(new SendResponseAction(channelId, reply));
	}
	
	
	/**
	 * Put a single {@link Action} into the queue to send a renewSessionReceived
	 * response on the channel identified by channelId.
	 * 
	 * @param channelId
	 */
	private void respondWithRenewSessionReceived(ChannelIdentifier channelId) {
		InputStream reply = mResponseCreator.createRenewSessionResponse();
		putIntoQueue(new SendResponseAction(channelId, reply));
	}

	/**
	 * Simple helper to put a {@link ActionSeries} containing only one
	 * {@link Action} object into the queue.
	 * 
	 * Hides creation of a {@link ActionSeries}.
	 * 
	 * @param singleAction
	 */
	private void putIntoQueue(Action singleAction) {
		putActionSeriesIntoActionQueue(new ActionSeries(singleAction));
	}
	
	/**
	 * Helper to put a {@link ActionSeries} into the action queue.
	 * 
	 * This is the point where we order the {@link ActionProcessor} on the
	 * other side to do something for us.
	 * 
	 * TODO: If we want to shutdown cleanly, we should probably honor
	 * a {@link InterruptedException} and the {@link Thread#isInterrupted()}
	 * further use the method?
	 * 
	 * @param as the actions to be put into the queue
	 */
	private void putActionSeriesIntoActionQueue(ActionSeries as) {
		NullCheck.check(as, "actionseries is null");
		boolean added = false;
		do {
			try {
				mActionQueue.put(as);
				added = true;
			} catch (InterruptedException e) { }
		} while (!added);
	}
	

	
	
	/**********************************************************************
	 * A bunch of getters and setters                                     *
	 **********************************************************************/	
	
	public void setActionQueue(Queue<ActionSeries> actionQueue) {
		NullCheck.check(actionQueue, "param actionQueue null");
		mActionQueue = actionQueue;
	}

	public void setPublisherIdProv(PublisherIdProvider publisherIdProv) {
		NullCheck.check(publisherIdProv, "param publisherIdProv null");
		mPublisherIdProv = publisherIdProv;
	}

	public void setSessionIdProv(SessionIdProvider sessionIdProv) {
		NullCheck.check(sessionIdProv, "param sessionIdProv null");
		mSessionIdProv = sessionIdProv;
	}

	public void setRequestTransformer(RequestTransformer requestTransformer) {
		NullCheck.check(requestTransformer, "param requestTransformer null");
		mRequestTransformer = requestTransformer;
	}

	public void setResponseCreator(ResponseCreator responseCreator) {
		NullCheck.check(responseCreator, "param responseCreator null");
		mResponseCreator = responseCreator;
	}

	public void setServerConfiguration(ServerConfigurationProvider serverConf) {
		NullCheck.check(serverConf, "param serverConf null");
		mServerConf = serverConf;
	}
	
	public void setDataModel(DataModelService dataModel) {
		NullCheck.check(dataModel, "param dataModel null");
		mDataModel = dataModel;
	}

	public void setSessionTimerFactory(SessionTimerFactory timerfactory) {
		NullCheck.check(timerfactory, "param timerfactory null");
		mSessionTimerFactory = timerfactory;
	}


	public void setPublisherIdGenerator(PublisherIdGenerator pubIdGen) {
		NullCheck.check(pubIdGen, "param pubIdGen null");
		mPublisherIdGenerator = pubIdGen;
	}
	
	public void setAuthorizationProv(AuthorizationProvider authProv) {
		NullCheck.check(authProv, "authProv null");
		mAuthorizationProv = authProv;
	}
}
