/*
 * Project: irond
 * Package: src.de.fhhannover.inform.iron.mapserver.datamodel
 * File:    ClientService.java
 *
 * Copyright (C) 2010-2011 Fachhochschule Hannover
 * Ricklinger Stadtweg 118, 30459 Hannover, Germany 
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.fhhannover.inform.iron.mapserver.datamodel;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.log4j.Logger;

import de.fhhannover.inform.iron.mapserver.datamodel.exceptions.NoSuchPublisherException;
import de.fhhannover.inform.iron.mapserver.datamodel.exceptions.PublisherConstructionException;
import de.fhhannover.inform.iron.mapserver.datamodel.exceptions.PurgePublisherNoAllowedException;
import de.fhhannover.inform.iron.mapserver.datamodel.exceptions.RunningSessionException;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.meta.Metadata;
import de.fhhannover.inform.iron.mapserver.logging.Logging;

/**
 * This class is responsible to handle newSession, endSession and purgePublisher.
 * DataModelService simply delegates the above calls to newSession(), endSession()
 * and purgePublish() respectively.
 * 
 * @author aw
 * @version 0.1
 */

/*
 * created: 05.12.09
 * changes:
 * 	05.12.09 aw - First version of this class
 *  17.12.09 aw - Return a NewSessionResult
 *  17.12.09 aw - Check whether a session is already there.
 *  04.02.10 aw - added logging with log4j
 *  05.02.10 aw - Use ResponseFactory
 *  07.02.10 aw - Fixed initialization
 *  10.02.10 aw - Moved purgePublisher() in here
 *  11.02.10 vp - the visibility of the classes and methods is changed
 *  18.02.10 vp - purgePublisher return PurgePublisherReceived or ErrorResult and check 
 *  	whether mapclient is allowed to do purgePublish
 *  27.04.10 aw - cleaned up methods, rather throw exceptions when errors
 *  	occur instead of returning "error responses" which might not be checked
 *  30.04.10 aw - remove singletons!
 *  30.05.10 aw - remove subscriptions when endSession() is called
 *  02.12.10 aw - Simplify endSession, newSession, purgePublisher,
 *                no complex parameters anymore
 */
class ClientService {
	
	/**
	 * static logger instance
	 */
	private static Logger logger;
	static {
		logger = Logging.getTheLogger();
	}

	private PublisherRep publisherRep;
	private MetaController metaController;
	private SubscriptionService subService;
	
	ClientService(PublisherRep pr, MetaController mCtl, SubscriptionService subServ) {
		publisherRep = pr;
		metaController = mCtl;
		subService = subServ;
	}
	
	/**
	 * Method which handles newSession calls.
	 * Simply a new publisher is added to the publisherRepository.
	 * In case of errors exceptions are thrown.
	 * 
	 * @param request
	 * @throws RunningSessionException newSession is called, the publisher exists
	 * 			but had never received an end session call
	 * @throws PublisherConstructionException misc exception if the request is bad
	 */
	void newSession(String sessionId, String publisherId) throws RunningSessionException,
												PublisherConstructionException {
		logger.trace("NewSession: sessionid=" + sessionId +
				" publisherid=" + publisherId);
	
		publisherRep.addPublisher(publisherId, sessionId);
	}
	 
	/**
	 * Method which handles an endSession call.
	 * First check whether the publisher is available, if not
	 * throw a NoSuchPublisherException.
	 * 
	 * Delete all session metadata from the publisher with the help
	 * of MetaController
	 * 
	 * Remove the session from the publisher and the corresponding
	 * session in the PublisherRep.
	 * 
	 * 
	 * @param request
	 * @return
	 * @throws NoSuchPublisherException if the publisher does not exists
	 * @throws ResponseCreationException 
	 */
	void endSession(String sessionId) throws NoSuchPublisherException {
		Map<Integer, Metadata> changes = new HashMap<Integer, Metadata>();
		logger.debug("EndSession: sessionid=" + sessionId);
	
		// throws exception if no publisher found
		Publisher p = publisherRep.getPublisherBySessionId(sessionId);

		// drop all subscriptions
		removeSubscriptions(sessionId);
		
		// delete session metadata of this publisher
		List<Metadata> toremove = p.getSessionMetadata();
		logger.trace("Removing " + toremove.size() + " session metadata objects");
		metaController.deleteMetadata(p, toremove, changes);
		
		// remove the session
		logger.trace("Removing session from " + p);
		p.deleteSessionId();
		publisherRep.removePublisherSession(sessionId);
		
		subService.updateSubscriptions(changes);
		
	}
	
	/**
	 * Remove all active subscriptions of the publisher with the given
	 * session-id.
	 * 
	 * @param sessionId
	 * @throws NoSuchPublisherException
	 */
	void removeSubscriptions(String sessionId) throws NoSuchPublisherException {
		Publisher p = publisherRep.getPublisherBySessionId(sessionId);
		
		for (Subscription sub : p.getSubscriptions().values()) {
			logger.trace("Clearing containers of subscription \"" + sub.getName() + "\"");
			subService.clearContainers(sub);
		}
		p.setNotified(false);
		p.clearSubscrptions();
	}
	
	/**
	 * Remove all metadata from publisher
	 * 
	 * @param request
	 * @return
	 * @throws NoSuchPublisherException 
	 * @throws PurgePublisherNoAllowedException 
	 * @throws ResponseCreationException 
	 */
	void purgePublisher(String sessionId, String publisherId) throws NoSuchPublisherException,
												PurgePublisherNoAllowedException {
		Map<Integer, Metadata> changes = new HashMap<Integer, Metadata>();
		if (DataModelService.getServerConfiguration().getPurgePublisherIsRestricted()) {
			Publisher publisher = publisherRep.getPublisherBySessionId(sessionId);
			
			if (!publisher.getPublisherId().equals(publisherId)) {
				logger.warn("purgePublisher not allowed");
				throw new PurgePublisherNoAllowedException("Not allowed to" +
						" purge metadata! of a different publisher");
			}
		}
		Publisher publisher = publisherRep.getPublisherByPublisherId(publisherId);
		logger.debug("Purging metadata of " + publisher);

		
		List<Metadata> mdlist = publisher.getSessionMetadata();
		logger.trace("Deleting n=" + mdlist.size() + " session metadata objects");
		metaController.deleteMetadata(publisher, mdlist, changes);
		
		mdlist = publisher.getLifetimeMetadata();
		logger.trace("Deleting n=" + mdlist.size() + " lifetime metadata objects");
		metaController.deleteMetadata(publisher, mdlist, changes);
		
		subService.updateSubscriptions(changes);
	}
}
