/*
 * Project: irond
 * Package: src.de.fhhannover.inform.iron.mapserver.datamodel
 * File:    DataModelService.java
 *
 * Copyright (C) 2010-2011 Fachhochschule Hannover
 * Ricklinger Stadtweg 118, 30459 Hannover, Germany 
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.fhhannover.inform.iron.mapserver.datamodel;

import de.fhhannover.inform.iron.mapserver.datamodel.exceptions.AlreadyObservedException;
import de.fhhannover.inform.iron.mapserver.datamodel.exceptions.InvalidMetadataException;
import de.fhhannover.inform.iron.mapserver.datamodel.exceptions.NoPollResultAvailableException;
import de.fhhannover.inform.iron.mapserver.datamodel.exceptions.NoSuchPublisherException;
import de.fhhannover.inform.iron.mapserver.datamodel.exceptions.NoSuchSubscriptionException;
import de.fhhannover.inform.iron.mapserver.datamodel.exceptions.ParameterException;
import de.fhhannover.inform.iron.mapserver.datamodel.exceptions.PublisherConstructionException;
import de.fhhannover.inform.iron.mapserver.datamodel.exceptions.PurgePublisherNoAllowedException;
import de.fhhannover.inform.iron.mapserver.datamodel.exceptions.ResponseCreationException;
import de.fhhannover.inform.iron.mapserver.datamodel.exceptions.RunningSessionException;
import de.fhhannover.inform.iron.mapserver.messages.DumpResult;
import de.fhhannover.inform.iron.mapserver.messages.PollResult;
import de.fhhannover.inform.iron.mapserver.messages.PublishRequest;
import de.fhhannover.inform.iron.mapserver.messages.SearchRequest;
import de.fhhannover.inform.iron.mapserver.messages.SearchResult;
import de.fhhannover.inform.iron.mapserver.messages.SubscribeRequest;
import de.fhhannover.inform.iron.mapserver.provider.DataModelServerConfigurationProvider;
import de.fhhannover.inform.iron.mapserver.utils.NullCheck;

/**
 * Entry Points for all Operations.
 * 
 * This class delegates all calls to the corresponding service objects.
 * 
 * @author awe, vp
 * @version 0.1
 * 
 */

/*
 * created: 19.11.09
 * changes:
 * 	22.11.09 vp - implemented publish()
 * 	26.11.09 aw - created singleton
 *  09.12.09 aw - search() call forwarded to SearchService,
 *                set version to 0.1
 *  04.02.10 aw - Added log4j logger, converted System.err.println() calls
 *  07.02.10 aw - Add checks for null requests, throw exception
 *  08.02.10 aw - Refactoring of constructor.
 *  10.02.10 aw - removed configureLogging()
 *  30.04.10 aw - created newInstance(), remove Singleton calls, 
 *                forwarding of subscription requests, removed logger as it
 *                isn't used here anyway, remove singletons!
 *  12.06.10 aw - make everything synchronized
 *  19.08.10 aw - some cleanup
 *  02.12.10 aw - Simplify endSession, newSession and purgePublisher, no complex
 *                parameters anymore, only sessionid (and publisherid) are needed

 */
public class DataModelService implements SubscriptionNotifier {
	
	/**
	 * this static fields represents the current serverConfiguration to
	 * be used.
	 * Necessary for the case sensitive settings and the purge publisher
	 * operation.
	 * 
	 * FIXME: This is currently overridden by each newInstance() call.
	 * Not really what we want, but I don't see a way around this right now.
	 */
	private static DataModelServerConfigurationProvider sServerConfiguration;
	
	private PublishService publishService;
	private ClientService clientService;
	private SearchService searchService;
	private SubscriptionService	subscriptionService;
	private PublisherRep publisherRep;
	private IdentifierRep identifierRep;
	private LinkRep linkRep;
	private MetaController metaController;
	
	
	private DataModelService() {
	
		publisherRep = new PublisherRep();
		
		identifierRep = new IdentifierRep();
		
		linkRep	= new LinkRep(identifierRep);
		
		metaController = new MetaController(identifierRep, linkRep);
		
		searchService = new SearchService(identifierRep, publisherRep);
		
		subscriptionService = new SubscriptionService(identifierRep,
													  publisherRep,
													  metaController);
		
		publishService = new PublishService(publisherRep, identifierRep,
											linkRep, metaController,
											subscriptionService);
		
		clientService = new ClientService(publisherRep, metaController,
										  subscriptionService);
	}
 
	
	/**
	 * Returns a <b>new </b> {@link DataModelService} instance.
	 * 
	 * Note:
	 * Calling this twice gives two completely independent DataModels. Might
	 * be confusing but it's nice at the same time.
	 * 
	 * @return a <b>new</b> instance of a {@link DataModelService} instance
	 */
	public static DataModelService newInstance(DataModelServerConfigurationProvider serverConf) {
		NullCheck.check(serverConf, "serverConf is null");
		sServerConfiguration = serverConf;
		
		return new DataModelService();
	}
	
	public static DataModelServerConfigurationProvider getServerConfiguration() {
		return sServerConfiguration;
	}
	
	/**
	 * A operation of publish is for create, modify or delete the metadata by
	 * one or more identifiers or links.
	 *
	 * @param PublishRequest
	 * @return publishReceived
	 * @throws ParameterException 
	 * @throws NotSupportedException 
	 * @throws NoSuchPublisherException
	 * @throws InvalidMetadataException 
	 * @throws ResponseCreationException 
	 */
	synchronized public void publish(PublishRequest request)
		throws NoSuchPublisherException, InvalidMetadataException {
		checkNull(request);
		publishService.publish(request);
	}


	synchronized public SearchResult search(SearchRequest request)
		throws NoSuchPublisherException {
		checkNull(request);
		return searchService.search(request);
	}
	 
	synchronized public void purgePublisher(String sessionId, String publisherId)
		throws NoSuchPublisherException, PurgePublisherNoAllowedException {
		checkNull(sessionId);
		checkNull(publisherId);
		clientService.purgePublisher(sessionId, publisherId);
	}
	
	synchronized public void newSession(String sessionId, String publisherId)
		throws RunningSessionException, PublisherConstructionException {
		checkNull(sessionId);
		checkNull(publisherId);
		clientService.newSession(sessionId, publisherId);
	}
	 
	synchronized public void endSession(String sessionId)
		throws NoSuchPublisherException {
		checkNull(sessionId);
		clientService.endSession(sessionId);
	}
	
	/* 
	 * Special Operations for visulaization
	 */
	synchronized public DumpResult dump(String sessionId) {
		checkNull(sessionId);
		DumpResult result = new DumpResult(sessionId);
		result.setIdentifier(identifierRep.getAllIdentifiers());
		result.setLastUpdateTime(identifierRep.getTimestamp());
		return result;
	}
	
	synchronized public long getLastUpdateTimestamp(){
		return identifierRep.getTimestamp();
	}
	
	synchronized public void subscribe(SubscribeRequest request)
		throws NoSuchPublisherException, NoSuchSubscriptionException {
		checkNull(request);
		subscriptionService.subscribe(request);
	}

	
	/**
	 * Debug graphviz output!!!
	 * @return
	 *
	synchronized public String graphViz() {
		StringBuffer sb = new StringBuffer();
		sb.append("graph {\n");
		sb.append(identifierRep.graphvizNodes());
		sb.append(linkRep.graphvizEdges());
		sb.append("\n\tlabel=\"IF-MAP Graph\"\n");
		sb.append("}");
		return sb.toString();
	}
	*/


	@Override
	synchronized public PollResult getPollResultFor(String sessionId)
			throws NoPollResultAvailableException, NoSuchPublisherException {
		checkNull(sessionId);
		return subscriptionService.getPollResultFor(sessionId);
	}


	@Override
	synchronized public void registerSubscriptionObserver(SubscriptionObserver subObs)
			throws AlreadyObservedException {
		checkNull(subObs);
		subscriptionService.setSubscriptionObserver(subObs);
	}
	
	private void checkNull(Object obj) throws NullPointerException {
		if (obj == null) {
			throw new NullPointerException("null was given");
		}
	}


	public void removeSubscriptions(String sessionId) throws NoSuchPublisherException {
		checkNull(sessionId);
		clientService.removeSubscriptions(sessionId);
	}
}
