/*
 * Project: irond
 * Package: src.de.fhhannover.inform.iron.mapserver.datamodel
 * File:    SearchResultImpl.java
 *
 * Copyright (C) 2010-2011 Fachhochschule Hannover
 * Ricklinger Stadtweg 118, 30459 Hannover, Germany 
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.fhhannover.inform.iron.mapserver.datamodel;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

import de.fhhannover.inform.iron.mapserver.datamodel.graph.Identifier;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.Link;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.MetadataContainer;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.meta.Metadata;
import de.fhhannover.inform.iron.mapserver.messages.PollResult;
import de.fhhannover.inform.iron.mapserver.messages.SearchResult;

/**
 * A new implementation of SearchResult which works with
 * two lists of Identifiers and Links.
 * Every Identifier|Link has another list which contains
 * the Metadata of this identifier.
 * 
 * In comparison to SearchResult1 the functionality of add()
 * and remove() Metadata objects from the MetadataContainer class
 * is not used rather it is stored somewhere else. Makes much more
 * sense for {@link PollResult}
 * 
 * @version 0.1
 * @author aw
 * 
 */

/*
 * created: 30.04.10 aw
 * 
 * changes:
 * 	30.04.10 aw - created class
 *  01.05.10 aw - include optional name for subscription names,
 *                create a SearchResult superclass
 */
public class SearchResultImpl extends SearchResult {

	/**
	 * if used in a PollResult this name represents the name of the subscription.
	 */
	protected List<Identifier> identifiers;
	protected List<List<Metadata>> iMetadata;
	protected List<Link> links;
	protected List<List<Metadata>> lMetadata;
	
	/**
	 * Constructor with a name
	 * @param name
	 * @throws ResponseCreationException 
	 */
	public SearchResultImpl(String sessionId, String name) {
		super(sessionId, name);
			
		identifiers = new ArrayList<Identifier>();
		iMetadata = new ArrayList<List<Metadata>>();
		links = new ArrayList<Link>();
		lMetadata = new ArrayList<List<Metadata>>();
	}

	
	/**
	 * Constructor which sets the name to null
	 * @param sessionId
	 * @throws ResponseCreationException
	 */
	public SearchResultImpl(String sessionId) {
		this(sessionId, null);
	}
	
	/* (non-Javadoc)
	 * @see de.fhhannover.inform.iron.mapserver.messages.SearchResult#addMetadata(de.fhhannover.inform.iron.mapserver.datamodel.graph.MetadataContainer, de.fhhannover.inform.iron.mapserver.datamodel.graph.meta.Metadata)
	 */
	@Override
	public void addMetadata(MetadataContainer mc, Metadata m) {
		if (mc == null)
			return;
		
		List<Metadata> mcMetadata = null;
		if (mc instanceof Identifier) {
			Identifier ident = (Identifier)mc;
			int index = identifiers.indexOf(ident);
			if (index >= 0) {
				mcMetadata = iMetadata.get(index);
			} else {									// identifier not in list
				identifiers.add(ident.copy());
				mcMetadata = new LinkedList<Metadata>();
				iMetadata.add(mcMetadata);
			}
			
		} else if (mc instanceof Link) {
			Link link = (Link)mc;
			int index = links.indexOf(link);
			if (index >= 0) {
				mcMetadata = lMetadata.get(index);
			} else {									// link not in list
				Link newLink = new Link(link.getIdent1().copy(), link.getIdent2().copy());
				links.add(newLink);
				mcMetadata = new LinkedList<Metadata>();
				lMetadata.add(mcMetadata);
			}
		}
		if (m != null) {
			mcMetadata.add(m.copy());
		}
	}
	
	/* (non-Javadoc)
	 * @see de.fhhannover.inform.iron.mapserver.messages.SearchResult#addMetadataContainer(de.fhhannover.inform.iron.mapserver.datamodel.graph.MetadataContainer)
	 */
	@Override
	public void addMetadataContainer(MetadataContainer mc) {
		if (mc != null) {
			addMetadata(mc, null);
		}
	}
	
	/* (non-Javadoc)
	 * @see de.fhhannover.inform.iron.mapserver.messages.SearchResult#getIdentifiers()
	 */
	@Override
	public List<Identifier> getIdentifiers() {
		return identifiers;
	}

	/* (non-Javadoc)
	 * @see de.fhhannover.inform.iron.mapserver.messages.SearchResult#getLinks()
	 */
	@Override
	public List<Link> getLinks() {
		return links;
	}

	
	/* (non-Javadoc)
	 * @see de.fhhannover.inform.iron.mapserver.messages.SearchResult#getMetadataFor(de.fhhannover.inform.iron.mapserver.datamodel.graph.MetadataContainer)
	 */
	@Override
	public List<Metadata> getMetadataFor(MetadataContainer mc) {
		int index = -1;
		List<Metadata> ret = null;
		if (mc instanceof Identifier) {
			index = identifiers.indexOf(mc);
			if (index >= 0)
				ret = iMetadata.get(index);
			
		} else if (mc instanceof Link) {
			index = links.indexOf(mc);
			if (index >= 0) {
				ret = lMetadata.get(index);
			}
		}
		return ret;
	}
	
	

}
