/*
 * Project: irond
 * Package: src.de.fhhannover.inform.iron.mapserver.datamodel.graph
 * File:    Device.java
 *
 * Copyright (C) 2010-2011 Fachhochschule Hannover
 * Ricklinger Stadtweg 118, 30459 Hannover, Germany 
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.fhhannover.inform.iron.mapserver.datamodel.graph;

import de.fhhannover.inform.iron.mapserver.datamodel.exceptions.IdentifierConstructionException;

/**
 * Device
 * 
 * Represents a Device Identifier.
 * Type can either be name or aikName
 * 
 * @author aw
 * @version 0.1
 * 
 * created: 26.11.09
 * changes:
 *  26.11.09 aw - Created first version of this class
 *  27.11.09 aw - equals() check for null
 *  09.12.09 aw - implemented copy()
 *  17.12.09 aw - added type.
 *  04.02.10 aw - toString()
 *  07.02.10 aw - modified toString()
 */
public class Device extends Identifier {
	
	private final String value;
	private final DeviceTypeEnum type;
	private int hashCode;
	
	/**
	 * Constructor using type and value
	 * value.length has to be greater 0			//FIXME: Really?
	 * 
	 * @param value
	 * @param type
	 * @throws IdentifierConstructionException
	 */
	public Device(final String value, final DeviceTypeEnum type) throws IdentifierConstructionException {
		super(IdentifierType.instanceOfTypeWith("device"));
		if (value == null || type == null || value.length() == 0) {
			throw new IdentifierConstructionException("Device: " +
					"invalid value or type");
		}

		// Use of aik-name is deprecated
		if (type == DeviceTypeEnum.aikName)
			throw new IdentifierConstructionException("Device with aik-name is deprecated");
		
		
		this.value = value;
		this.type = type;
		this.hashCode = (super.hashCode() + type.toString() + value).hashCode();
	}
	
	
	public String getValue() {
		return value;
	}
	
	
	public DeviceTypeEnum getDeviceType() {
		return type;
	}
	
	/**
	 * Two devices are the same if
	 *  - they have the same reference
	 *  - they have the same type and the same value
	 */ 
	@Override
	public boolean equals(Object o) {
		Device d;
		
		if (o == null)
			return false;
		
		if (this == o)
			return true;
		
		if (o instanceof Device) {
			d = (Device) o;
			if (type == d.getDeviceType() &&
					value.equals(d.getValue()))
				return true;
		}
		return false;
	}
	
	public int hashCode() {
		return hashCode;
	}


	@Override
	public Identifier copy() {
		Device device = null;
		try {
			device = new Device(this.getValue(), this.getDeviceType());
		} catch (IdentifierConstructionException e) {
			e.printStackTrace();					/* should not fail */
		}
		return device;
	}
	
	public String toString() {
		return "Device<value=" + getValue() + " type=" +
			type.toString() +">";
	}
}
 
