/*
 * Project: irond
 * Package: src.de.fhhannover.inform.iron.mapserver.datamodel.graph
 * File:    Identity.java
 *
 * Copyright (C) 2010-2011 Fachhochschule Hannover
 * Ricklinger Stadtweg 118, 30459 Hannover, Germany 
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.fhhannover.inform.iron.mapserver.datamodel.graph;

import de.fhhannover.inform.iron.mapserver.datamodel.DataModelService;
import de.fhhannover.inform.iron.mapserver.datamodel.exceptions.IdentifierConstructionException;
import de.fhhannover.inform.iron.mapserver.utils.IpAddressValidator;

/**
 * Class which represents a Identity.
 * 
 * A Identity can be of different types which are represented
 * by a DeviceType.
 * 
 * 
 * @author aw
 * 
 * @version 0.1
 *
 * created: 26.11.09 
 * changes:
 * 	26.11.09 aw - Created first version of this class
 * 	27.11.09 aw - equals() check for null, added constructor
 *  	to be used with default administrative domain
 *  28.11.09 aw - equals() use super.equals() to compare administrative
 * 		domains.
 *  09.12.09 aw - implemented copy()
 *  17.12.09 aw - added type
 *  04.02.10 aw - toString()
 *  07.02.10 aw - modified toString(), use ConfigurationManager with new
 *                type based function to determine wheter case sensitivity is
 *                needed.
 *  12.02.10 aw - modified to use new IdentityType enum
 *  10.10.10 aw - copy() needs to check if the type is other because then a
 *                different constructor needs to be used
 *  
 *  
 * what about:
 *  - fixing the equals method...
 *
 */
public class Identity extends  IdentifierWithAdministrativeDomain {
	
	private final String name;
	private IdentityTypeEnum type;
	private String otherTypeDefinition;
	private int hashCode;
	
	/**
	 * Private constructor that mostly does error checking.
	 * 
	 * Checks name for null and checks length > 0
	 * 
	 * @param name
	 * @param ad
	 * @throws IdentifierConstructionException
	 */
	private Identity(final String name, final String ad) throws IdentifierConstructionException {
		super(IdentifierType.instanceOfTypeWith("identity"), ad);
		if (name == null || name.length() == 0)
			throw new IdentifierConstructionException("Identity: name is invalid");
		this.name = name;
	}
	
	/**
	 * Construct a identity object from name, administrative-domain and the type.
	 * 
	 * *Note:* If we come across a Host Identity Tag (hipHit), treat it as IPv6
	 * address.
	 * 
	 * Using this constructor it is not allowed to
	 * set it to IdentityType.other!
	 * 
	 * @param name
	 * @param ad
	 * @param it
	 * @throws IdentifierConstructionException 
	 */
	public Identity(final String name, final String ad, final IdentityTypeEnum it) throws IdentifierConstructionException {
		this(name, ad);
		if (it == null)
			throw new IdentifierConstructionException("Identity: type is null");
		if (it == IdentityTypeEnum.other)
			throw new IdentifierConstructionException("Identity: for type other" +
					"use the other constructor!");
		
		if (it == IdentityTypeEnum.hipHit) {
			if (!IpAddressValidator.validateIPv6(this.getName())) {
				throw new IdentifierConstructionException("HIP-HIT incorrect format");
			}
		}
		type = it;
		String hashString = super.hashCode() + this.type.toString() + this.name;
		hashCode = hashString.hashCode();
	}
	
	/**
	 * Construct a identity object from name,
	 * administrative-domain and a other-type-definition.
	 * 
	 * @param name
	 * @param ad
	 * @param otherTypeDef
	 * @throws IdentifierConstructionException
	 */
	public Identity(final String name, final String ad, final String otherTypeDef) throws IdentifierConstructionException {
		this(name, ad);
		if (otherTypeDef == null)
			throw new IdentifierConstructionException("otherTypeDef " +
					"is not allowed to be null!");
		this.otherTypeDefinition = otherTypeDef;
		this.type = IdentityTypeEnum.other;
		String hashString = super.hashCode() + type.toString();
		hashString += this.otherTypeDefinition;
		hashString += this.name;
		hashCode = hashString.hashCode();
	}
	
	/**
	 * Constructor for Identity Identifier to create an empty
	 * administrativeDomain with a type other than other.
	 * 
	 * @param name
	 * @param it
	 * @throws IdentifierConstructionException
	 */
	public Identity(String name, IdentityTypeEnum it) throws IdentifierConstructionException {
		this(name, "", it);
	}
	
	
	public int hashCode() {
		return hashCode;
	}
	
	public IdentityTypeEnum getIdentityType() {
		return type;
	}
	
	public String getName() {
		return name;
	}
	
	/**
	 * Might return null if the type of this Identity is not "other"
	 * 
	 * @return
	 */
	public String getOtherTypeDefinition() {
		if (type != IdentityTypeEnum.other)
			return null;
		
		return otherTypeDefinition;
	}

	/**
	 * handle every case
	 *  - same reference
	 *  - same type is needed!
	 *  
	 *  FIXME: THIS IS COMPLETLY UGLY
	 */
	@Override
	public boolean equals(Object o) {
		boolean caseSensitive = false;
		boolean nameRes = false;
		Identity i;
		IdentityTypeEnum it;
		String iname;
		String iotherTypeDef;
		
		if (this == o)
			return true;
		
		if (super.equals(o)) {
			if (o instanceof Identity) {
				i = (Identity) o;
				it = i.getIdentityType();
				iname = i.getName();
			
				// same type ?
				if (type != it)
					return false;
			
				// request whether this type has to be checked case sensitive
				caseSensitive = DataModelService.getServerConfiguration().getIdentityTypeIsCaseSensitive(type.toString());
				
				// compare other-type-definitions
				if (type == IdentityTypeEnum.other) {
					iotherTypeDef = i.getOtherTypeDefinition();
					
					// we check the other type def case sensitive... ?
					// FIXME: What is the specification saying?
					if (!otherTypeDefinition.equals(iotherTypeDef))		
						return false;
				}
				if (caseSensitive) {
					nameRes = name.equals(iname);
				} else {
					nameRes = name.equalsIgnoreCase(iname);
				}
			}
		}
		return nameRes;
	}

	@Override
	public Identifier copy() {
		Identity id = null;
		
		try {
			if (this.getIdentityType() == IdentityTypeEnum.other) {
				id = new Identity(this.getName(), this.getAdministrativeDomain(),
						this.getOtherTypeDefinition());
			} else {
				id = new Identity(this.getName(), this.getAdministrativeDomain(),
															this.getIdentityType());
			}
				
		} catch (IdentifierConstructionException e) {
			e.printStackTrace();						/* should not fail */
			// But it did (10.10.2010 aw)
		}
		
		return id;
	}
	
	public String toString() {
		String ad = getAdministrativeDomain();
		StringBuffer sb = new StringBuffer();
		sb.append("Identity<name=" + name);
		if (!(ad == null | ad.length() == 0)) {
			sb.append(" ad=" + ad);
		}
		sb.append(" type=" + type.toString());
		if (type == IdentityTypeEnum.other) {
			sb.append(" def=" + otherTypeDefinition);
		}
		sb.append(">");
		return sb.toString();
	}
}