/*
 * Project: irond
 * Package: src.de.fhhannover.inform.iron.mapserver.datamodel.graph
 * File:    IpAddress.java
 *
 * Copyright (C) 2010-2011 Fachhochschule Hannover
 * Ricklinger Stadtweg 118, 30459 Hannover, Germany 
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.fhhannover.inform.iron.mapserver.datamodel.graph;

import de.fhhannover.inform.iron.mapserver.datamodel.exceptions.IdentifierConstructionException;
import de.fhhannover.inform.iron.mapserver.utils.IpAddressValidator;

/**
 * IpAddress Identifier Type
 * 
 * @author aw
 * @version 0.1
 * 
 * created: 26.11.09
 * changes:
 *  26.11.09 aw - Created first version of this class
 *  27.11.09 aw - Implemented equals(), created constructor with
 *  	default administrative-domain
 *  28.11.09 aw - equals() use super.equals() to compare administrative
 * 		domains.
 *  09.12.09 aw - implemented copy()
 *  17.12.09 aw - added type
 *  04.02.10 aw - added toString()
 *  07.02.10 aw - modified toString()
 *  08.02.10 aw - added hashCode()
 */
public class IpAddress extends IdentifierWithAdministrativeDomain {
	
	private final String value;
	private final IpAddressTypeEnum type;
	private final int hashCode;
	
	/**
	 * Construct a IpAddress from value, administrative-domain
	 * and the type.
	 * 
	 * @param value
	 * @param ad
	 * @param ipat
	 * @throws IdentifierConstructionException
	 */
	public IpAddress(final String value, final String ad, IpAddressTypeEnum ipat) throws IdentifierConstructionException {
		super(IdentifierType.instanceOfTypeWith("ip-address"), ad);
		if (ipat == null)
			ipat = IpAddressTypeEnum.IPv4;
		
		if (value == null) 
			throw new IdentifierConstructionException("value " +
					"and type of IPAddress need to be set");
		
		if (ipat == IpAddressTypeEnum.IPv4) {
			if (!IpAddressValidator.validateIPv4(value)) {
				throw new IdentifierConstructionException(" Bad " +
						"IPv4 Format");
			}
		} else if (ipat == IpAddressTypeEnum.IPv6) {
			if (!IpAddressValidator.validateIPv6(value)) {
				throw new IdentifierConstructionException(" Bad " +
				"IPv6 Format");
			}
		}
		
		this.value = value;
		this.type = ipat;
		
		hashCode = (super.hashCode() + "" + this.type + this.value).hashCode();
	}
	
	public IpAddress(final String value, final IpAddressTypeEnum ipat) throws IdentifierConstructionException {
		this(value, "", ipat);
		
	}

	public String getValue() {
		return value;
	}

	public IpAddressTypeEnum getIpAddressType() {
		return type;
	}

	/**
	 * IP Address Identifier are equal if they have the same
	 * type and the same value;
	 * 
	 * The same value is case insensitive:
	 *  - IPv4 only contains digits
	 *  - IPv6 can contain hexadecimal characters
	 *    which may be upper or lowercase or digits.
	 * 
	 */
	@Override
	public boolean equals(Object o) {
		boolean valRes = false;
		IpAddress oip;

		if (this == o)
			return true;
		
		if (super.equals(o)) {
			if (o instanceof IpAddress) {
				oip = (IpAddress) o;
				if (type == oip.getIpAddressType()) {
					valRes = value.equals(oip.getValue());
				}
			}
		}
		return valRes;
	}

	@Override
	public Identifier copy() {
		IpAddress ip = null;
		try {
			ip = new IpAddress(this.getValue(), this.getAdministrativeDomain(),
													this.getIpAddressType());
		} catch (IdentifierConstructionException e) {
			e.printStackTrace();				/* should not fail */
		}
		return ip;
	}

	@Override
	public String toString() {
		String ad = getAdministrativeDomain();
		StringBuffer sb = new StringBuffer();
		sb.append("IP-Address<value=" + value);
		if (!(ad == null | ad.length() == 0)) {
			sb.append(" ad=" + ad);
		}
		sb.append(" type=" + type.toString() + ">");
		return sb.toString();
	}
	
	public int hashCode() {
		return hashCode;
	}
}
 
