/*
 * Project: irond
 * Package: src.de.fhhannover.inform.iron.mapserver.datamodel.graph.meta
 * File:    W3cXmlMetadata.java
 *
 * Copyright (C) 2010-2011 Fachhochschule Hannover
 * Ricklinger Stadtweg 118, 30459 Hannover, Germany 
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.fhhannover.inform.iron.mapserver.datamodel.graph.meta;


import java.util.Map;

import javax.xml.namespace.NamespaceContext;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.TransformerFactoryConfigurationError;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpression;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import de.fhhannover.inform.iron.mapserver.datamodel.Filter;
import de.fhhannover.inform.iron.mapserver.datamodel.exceptions.MetadataConstructionException;
import de.fhhannover.inform.iron.mapserver.utils.FilterAdaption;
import de.fhhannover.inform.iron.mapserver.utils.NullCheck;
import de.fhhannover.inform.iron.mapserver.utils.SimpleNamespaceContext;

public class W3cXmlMetadata extends Metadata {
	
	private Document xmlDocument;
	private Element xmlElement;
	private static XPathFactory xpathfac;
	private static DocumentBuilderFactory documentBuilderFactory;
	private static DocumentBuilder documentBuilder;

	// static initialization code for this class.
	{
		documentBuilderFactory = DocumentBuilderFactory.newInstance();
		documentBuilderFactory.setNamespaceAware(true);
		try {
			documentBuilder = documentBuilderFactory.newDocumentBuilder();
		} catch (ParserConfigurationException e) {
			logger.error("Could not get a DocumentBuilder instance: " + e.getMessage());
		}
		xpathfac = XPathFactory.newInstance();
	}

	public W3cXmlMetadata(Element node, MetaLifeTimeType lifetime) throws MetadataConstructionException {
		// As dangerous as before :-(
		super(MetaType.instanceOfTypeWith(
				node.getNamespaceURI() +":" + node.getLocalName(), node.getAttribute(CARDINALITYSTRING)), lifetime);
		NullCheck.check(node, "node is null");
		if (node.getNodeType() != Node.ELEMENT_NODE) {
			throw new MetadataConstructionException("node is not element node");
		}
		// we checked that before, didn't we?
		// import the node to a new document
		xmlDocument = documentBuilder.newDocument();
		Node cpyNode = xmlDocument.importNode(node, true);
		xmlElement = (Element)cpyNode;
		xmlDocument.appendChild(xmlElement);
	}
	
	/**
	 * FIXME: This might crash :-(
	 * Copy Constructor
	 * @param toCopy
	 * @throws MetadataConstructionException 
	 */
	private W3cXmlMetadata(W3cXmlMetadata toCopy) throws MetadataConstructionException {
		super(toCopy);
		xmlDocument = (Document) toCopy.xmlDocument.cloneNode(true);
		xmlElement = (Element) xmlDocument.getChildNodes().item(0);
	}
		
	

	@Override
	public boolean matchesFilter(Filter f) {
		NullCheck.check(f, "filter is null");
		/*
		logger.trace("matching with filter " + f.toString());
		*/

		
		// shortcut
		if (f.isMatchEverything()) {
			return true;
		}
	
		// shortcut
		if (f.isMatchNothing()) {
			return false;
		}
		
		String fs = f.getFilterString();
		XPath xpath = xpathfac.newXPath();
		
		Map<String, String> nsMap = f.getNamespaceMap();
		
		/*
		if (logger.isTraceEnabled()) {
			int cnt = 1;
			logger.trace("Namespace map used for matching:");
			for (Entry<String, String> e : nsMap.entrySet()) {
				logger.trace(cnt++ + ":\t" +  e.getKey() + " -- " + e.getValue());
			}
		}
		*/
		
		NamespaceContext nsCtx = new SimpleNamespaceContext(nsMap);
		xpath.setNamespaceContext(nsCtx);
	
		/*
		logger.trace("Filter before adaption: " + fs);
		*/

		// add * to lonely brackets
		fs = FilterAdaption.adaptFilterString(fs);
	
		/*
		logger.trace("Filter after adaption: " + fs);
		*/
		
		XPathExpression expr = null;

		// this should never happen, as we checked it before
		try {
			expr = xpath.compile(fs);
		} catch (XPathExpressionException e1) {
			logger.error("UNEXPECTED: Could not compile filterstring" + fs);
			return false;
		}
		
		Object ret = null;
		try {
			ret = expr.evaluate(xmlDocument, XPathConstants.BOOLEAN);
			/*
			logger.trace("matching result is " + ((Boolean)ret).booleanValue());
			*/
		} catch (XPathExpressionException e) {
			e.printStackTrace();
			logger.error("evaluate failed badly: " + e.getMessage());
			return false;
		}
		return ((Boolean)ret).booleanValue();
	}

	@Override
	public void setPublisherId(String pubid) {
		NullCheck.check(pubid, "publisher-id is null");
		if (pubid.length() == 0) {
			throw new RuntimeException("Fix your parameters, please...");
		}
		publisherId = pubid;
		xmlElement.setAttribute(PUBLISHERID, publisherId);
	}

	@Override
	public Document toW3cDocument() {
		return (Document) xmlDocument.cloneNode(true);
	}
	
	@Override
	public void setTimestamp(String ts) {
		NullCheck.check(ts, "timestamp is null");
		if (ts.length() == 0) {
			throw new RuntimeException("Fix your parameters, please...");
		}
		timestamp = ts;
		xmlElement.setAttribute(TIMESTAMP, timestamp);
	}

	@Override
	public Metadata copy() {
		Metadata ret = null;
		try {
			ret = new W3cXmlMetadata(this);
		} catch (MetadataConstructionException e) {
			// we constructed the metadata object before
			// so it should work a second time...
			e.printStackTrace();
		}
		return ret;
	}

	
	@SuppressWarnings("unused")
	private void outputXmlDocument() {
		Transformer trans = null;
		try {
			trans = TransformerFactory.newInstance().newTransformer();
		} catch (TransformerConfigurationException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		} catch (TransformerFactoryConfigurationError e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}
		Result result = new StreamResult(System.err);
		Source source = new DOMSource(xmlDocument);
		try {
			trans.transform(source, result);
		} catch (TransformerException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
}

