/*
 * Project: irond
 * Package: src.de.fhhannover.inform.iron.mapserver.messages
 * File:    SearchResult.java
 *
 * Copyright (C) 2010-2011 Fachhochschule Hannover
 * Ricklinger Stadtweg 118, 30459 Hannover, Germany 
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.fhhannover.inform.iron.mapserver.messages;

import java.util.ArrayList;
import java.util.List;

import de.fhhannover.inform.iron.mapserver.datamodel.graph.Identifier;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.Link;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.MetadataContainer;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.meta.Metadata;

/**
 * Abstract superclass which contains abstract methods that need to be
 * implemented by subclasses and further some logic that can probably be
 * reused.
 * 
 * @author aw
 * @version 0.1
 *
 */

/*
 * created: long time ago
 * changes:
 * 	   05.10 aw - extracting a abstract class
 *  08.06.10 aw - changed LinkedList to ArrayList
 */
public abstract class SearchResult extends Response {
	
	private String name;

	public SearchResult(String sessionId, String name) {
		super(sessionId);
		if (name != null && name.length() == 0)
			throw new RuntimeException(" name is set, so length should be > 0");
		if (name != null) {
			this.name = new String(name);
		}
	}
	
	
	/**
	 * Add metadata objects to the given metadata container. A implementation
	 * of {@link SearchResult} is responsible to make deep copies of both,
	 * {@link MetadataContainer} objects and {@link Metadata} objects.
	 * 
	 * @param mc
	 * @param m
	 */
	public abstract void addMetadata(MetadataContainer mc, Metadata m);
	
	/**
	 * Only add a MetadataContainer. A implementation of {@link SearchResult}
	 * is responsible to make a deep copy of the {@link MetadataContainer} object.
	 * @param mc
	 */
	public abstract void addMetadataContainer(MetadataContainer mc);

	/**
	 * @return a list of identifiers in this SearchResult, length may be zero
	 */
	public abstract List<Identifier> getIdentifiers();
	
	/**
	 * @return a list of links in the SearchResult, length may be zero
	 */
	public abstract List<Link> getLinks();
	
	
	/**
	 * Get a list of Metadata in this SearchResult by giving a
	 * Metadatacontainer as parameter.
	 * 
	 * @return a list of lists of Metadata for the given Metadatacontainer or
	 * 			null if this metadatacontainer is not in the search result.
	 */
	public abstract List<Metadata> getMetadataFor(MetadataContainer mc);
	
	
	/**
	 * Returns the name of the corresponding subscription or
	 * null if no name was given.
	 * 
	 * @return
	 */
	public String getName() {
		if (name != null)
			return new String(name);
		return null;
	}
	
	public String toString() {
		StringBuffer sb = new StringBuffer();
		sb.append("Identifiers:\n");
		for (Identifier i : getIdentifiers()) {
			sb.append("\t" + i.toString() + "\n");
			for (Metadata m : getMetadataFor(i)) {
				sb.append("\t\t" + m.getType().getTypeString() + "\n");
			}
		}
		sb.append("Links:\n");
		for (Link l : getLinks()) {
			sb.append("\t" + l.toString() + "\n");
			for (Metadata m : getMetadataFor(l)) {
				sb.append("\t\t" + m.getType().getTypeString() + "\n");
			}
		}
		return sb.toString();
	}


	public boolean isEmpty() {
		return (getIdentifiers().isEmpty() && getLinks().isEmpty());
	}
	
	public List<MetadataContainer> getContainers() {
		List<MetadataContainer> ret = new ArrayList<MetadataContainer>(getIdentifiers());
		ret.addAll(getLinks());
		return ret;
	}
}
