/*
 * Project: irond
 * Package: src.de.fhhannover.inform.iron.mapserver.utils
 * File:    JaxbIdentifierExtractor.java
 *
 * Copyright (C) 2010-2011 Fachhochschule Hannover
 * Ricklinger Stadtweg 118, 30459 Hannover, Germany 
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.fhhannover.inform.iron.mapserver.utils;


import java.util.List;

import org.trustedcomputinggroup._2010.ifmap._2.AccessRequestType;
import org.trustedcomputinggroup._2010.ifmap._2.DeleteType;
import org.trustedcomputinggroup._2010.ifmap._2.DeviceType;
import org.trustedcomputinggroup._2010.ifmap._2.IPAddressType;
import org.trustedcomputinggroup._2010.ifmap._2.IdentityType;
import org.trustedcomputinggroup._2010.ifmap._2.MACAddressType;
import org.trustedcomputinggroup._2010.ifmap._2.NotifyType;
import org.trustedcomputinggroup._2010.ifmap._2.SearchType;
import org.trustedcomputinggroup._2010.ifmap._2.UpdateType;

import de.fhhannover.inform.iron.mapserver.datamodel.exceptions.IdentifierConstructionException;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.Identifier;

/**
 * Extract Identifiers out of different request types...
 * 
 * @author aw, dl
 * 
 * created: 12.02.10
 * 
 * changes
 *  14.02.10 aw - removed redundant code
 * 
 */
class JaxbIdentifierExtractor {
	
	private static JaxbIdentifierExtractor instance;
	
	private JaxbIdentifierTransformer iTrans;

	static JaxbIdentifierExtractor getInstance() {
		
		if (instance == null)
			instance = new JaxbIdentifierExtractor();
		
		return instance;
	}
	
	private JaxbIdentifierExtractor() {
		iTrans = JaxbIdentifierTransformer.getInstance();
	}
	
	

	/**
	 * Extract a datamodel Identifier from a SearchType object and return
	 * it.
	 * 
	 * @param search
	 * @return
	 * @throws IdentifierConstructionException
	 */
	Identifier extractFromSearch(SearchType search)
										throws IdentifierConstructionException {
		Identifier ret = null;
		Object o = null;
		
		if (search != null) {
			if (search.getAccessRequest() != null)
				o = search.getAccessRequest();
			else if (search.getDevice() != null)
				o = search.getDevice();
			else if (search.getIdentity() != null)
				o = search.getIdentity();
			else if (search.getIpAddress() != null)
				o = search.getIpAddress();
			else if (search.getMacAddress() != null)
				o = search.getMacAddress();
			else
				throw new IdentifierConstructionException("identifier not set in search");
			
			ret = transformObjectToIdentifier(o);
		}
		return ret;
	}

	/**
	 * Return an array of datamodel Identifiers given a UpdateType
	 * 
	 * The returned array is of size 2 and might contain one element
	 * which is null.
	 * 
	 * @param value
	 * @return
	 * @throws IdentifierConstructionException
	 */
	Identifier[] extractFromUpdate(UpdateType value)
										throws IdentifierConstructionException {
		Identifier[] ret = null;
		if (value != null) {
			List<Object> list = value.getAccessRequestOrIdentityOrIpAddress();
			ret = new Identifier[2];
			
			for (int i = 0; i < list.size(); i++)
				ret[i] = transformObjectToIdentifier(list.get(i));
		}
		return ret;
	}


	/**
	 * Return an array of datamodel Identifiers given a DeleteType
	 * 
	 * The returned array is of size 2 and might contain one element
	 * which is null.
	 * 
	 * @param value
	 * @return
	 * @throws IdentifierConstructionException
	 */
	Identifier[] extractFromDelete(DeleteType value)
										throws IdentifierConstructionException {
		Identifier[] ret = null;
		if (value != null) {
			List<Object> list = value.getAccessRequestOrIdentityOrIpAddress();
			ret = new Identifier[2];
			for (int i = 0; i < list.size(); i++)
				ret[i] = transformObjectToIdentifier(list.get(i));
		}
		return ret;
	}
	
	Identifier[] extractFromNotify(NotifyType value) 
										throws IdentifierConstructionException {
		Identifier[] ret = null;
		if (value != null) {
			List<Object> list = value.getAccessRequestOrIdentityOrIpAddress();
			ret = new Identifier[2];
			for (int i = 0; i < list.size(); i++)
				ret[i] = transformObjectToIdentifier(list.get(i));
		}
		return ret;
	}
	
	
	/**
	 * Helper method which converts from a given Object o into
	 * a datamodel Identifier.
	 * 
	 * @param o
	 * @return
	 * @throws IdentifierConstructionException
	 */
	private Identifier transformObjectToIdentifier(Object o)
										throws IdentifierConstructionException {
		Identifier res = null;
		if (o != null) {
			if (o instanceof AccessRequestType) {
				res = iTrans.transformAR((AccessRequestType)o);
			} else if (o instanceof DeviceType) {
				res = iTrans.transformDevice((DeviceType)o);
			} else if (o instanceof IPAddressType) {
				res = iTrans.transformIP((IPAddressType)o);
			} else if (o instanceof IdentityType) {
				res = iTrans.transformIdentity((IdentityType)o);
			} else if (o instanceof MACAddressType) {
				res = iTrans.transformMacAddress((MACAddressType)o);
			} else {
				throw new IdentifierConstructionException("object not a identifier");
			}
		}
		return res;
	}
}
