/*
 * Project: irond
 * Package: src.de.fhhannover.inform.iron.mapserver.utils
 * File:    JaxbIdentifierTransformer.java
 *
 * Copyright (C) 2010-2011 Fachhochschule Hannover
 * Ricklinger Stadtweg 118, 30459 Hannover, Germany 
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.fhhannover.inform.iron.mapserver.utils;

import org.trustedcomputinggroup._2010.ifmap._2.AccessRequestType;
import org.trustedcomputinggroup._2010.ifmap._2.DeviceType;
import org.trustedcomputinggroup._2010.ifmap._2.IPAddressType;
import org.trustedcomputinggroup._2010.ifmap._2.IdentityType;
import org.trustedcomputinggroup._2010.ifmap._2.MACAddressType;

import de.fhhannover.inform.iron.mapserver.datamodel.exceptions.IdentifierConstructionException;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.AccessRequest;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.Device;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.DeviceTypeEnum;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.Identity;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.IdentityTypeEnum;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.IpAddress;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.IpAddressTypeEnum;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.MacAddress;

/**
 * This class contains methods which can be used to transform from autogenerated
 * Identifier types to datamodel Identifiers and back.
 * 
 * FIXME:
 * No guarantee for correct transformation, there really should be some tests...
 * 
 * FIXME:
 * Maybe we should extract an Interface out of this
 * 
 * @version 0.1
 * @author aw
 * 
 * created: 10.02.10
 * 
 * changes:
 *   12.02.10 aw - changed identifier transformation to "not" set the 
 *                 administrative domain if it's empty
 *               - Changing for jaxbri code
 *   14.02.10 aw - some cleanup and commenting, moded type enum transformations
 *                 into TypeTransformer class
 *
 */
class JaxbIdentifierTransformer {
	
	private static JaxbIdentifierTransformer instance = new JaxbIdentifierTransformer();
	
	private JaxbIdentityTypeTransformer typeTrans;
	
	
	/**
	 *  Private constructor
	 */
	private JaxbIdentifierTransformer() {
		typeTrans = JaxbIdentityTypeTransformer.getInstance();
	}
	
	
	/**
	 * getInstance() accessor
	 * 
	 * @return
	 */
	static JaxbIdentifierTransformer getInstance() {
		if (instance == null)
			instance = new JaxbIdentifierTransformer();
		
		return instance;
	}
	
	
	/**
	 * From autogenerated IP to datamodel IP
	 * 
	 * @return
	 * @throws ` 
	 */
	IpAddress transformIP(IPAddressType iat)
										throws IdentifierConstructionException {
		IpAddress ret = null;
		IpAddressTypeEnum ipt;
		if (iat != null) {
			ipt = typeTrans.transformIpAddressType(iat.getType());
			ret = new IpAddress(iat.getValue(), iat.getAdministrativeDomain(), ipt);
		}
		return ret;
	}

	
	/**
	 * From datamodel IP to autogenerated IP
	 * 
	 * @param ip
	 * @return
	 */
	IPAddressType transformIP(IpAddress ip) {
		IPAddressType ret = null;
		String type = null;
		if (ip != null) {
			ret = new IPAddressType();
			type = typeTrans.transformIpAddressType(ip.getIpAddressType());
			ret.setType(type);
			
			ret.setValue(ip.getValue());
			
			if (ip.getAdministrativeDomain().length() == 0) {
				ret.setAdministrativeDomain(null);
			} else {
				ret.setAdministrativeDomain(ip.getAdministrativeDomain());
			}
		}
		return ret;
	}

	
	/**
	 * From autogenerated Mac to datamodel Mac
	 * 
	 * @param mt
	 * @return
	 * @throws IdentifierConstructionException 
	 */
	MacAddress transformMacAddress(MACAddressType mt)
										throws IdentifierConstructionException {
		MacAddress ret = null;
		if (mt != null) {
			ret = new MacAddress(mt.getValue(), mt.getAdministrativeDomain());
		}
		return ret;
	}

	
	/**
	 * From datamodel Mac to autogenerated Mac
	 * 
	 * @param mac
	 * @return
	 */
	MACAddressType transformMacAddress(MacAddress mac) {
		MACAddressType ret = null;
		if (mac != null) {
			ret = new MACAddressType();
			ret.setValue(mac.getValue());
			if (mac.getAdministrativeDomain().length() == 0) {
				ret.setAdministrativeDomain(null);
			} else {
				ret.setAdministrativeDomain(mac.getAdministrativeDomain());
			}
		}
		return ret;
	}

	
	/**
	 * From autogenerated Identity to datamodel Identity
	 * 
	 * @param it
	 * @return
	 * @throws IdentifierConstructionException 
	 */
	Identity transformIdentity(IdentityType it)
										throws IdentifierConstructionException {
		Identity ret = null;
		IdentityTypeEnum type;
		if (it != null) {
			type = typeTrans.transformIdentityType(it.getType());
			if (type == IdentityTypeEnum.other) {
				ret = new Identity(it.getName(), it.getAdministrativeDomain(),
						it.getOtherTypeDefinition());
			} else {
				ret = new Identity(it.getName(), it.getAdministrativeDomain(),
						type);
			}
		}
		return ret;
	}

	
	/**
	 * From datamodel Identity to autogenerated Identity
	 * 
	 * @param identity
	 * @return
	 */
	IdentityType transformIdentity(Identity identity) {
		IdentityType ret = null;
		String t;
		if (identity != null) {
			ret = new IdentityType();
			t = typeTrans.transformIdentityType(identity.getIdentityType());
			ret.setType(t);
			ret.setName(identity.getName());
			
			if (identity.getAdministrativeDomain().length() == 0) {
				ret.setAdministrativeDomain(null);
			} else {
				ret.setAdministrativeDomain(identity.getAdministrativeDomain());
			}
			if (identity.getIdentityType() == IdentityTypeEnum.other) {
				ret.setOtherTypeDefinition(identity.getOtherTypeDefinition());
			}
		}
		return ret;
	}
	
	
	/**
	 * From autogenerated Device to datamodel Device
	 * 
	 * @param dt
	 * @return
	 * @throws IdentifierConstructionException
	 */
	Device transformDevice(DeviceType dt)
										throws IdentifierConstructionException {
		Device ret = null;
		if (dt != null) {
			if (dt.getAikName() != null) {
				ret = new Device(dt.getAikName(), DeviceTypeEnum.aikName);
			} else {
				ret = new Device(dt.getName(), DeviceTypeEnum.name);
			}
		}
		return ret;
	}

	
	/**
	 * From datamodel Device to autogenerated Device
	 * 
	 * @param dev
	 * @return
	 */
	DeviceType transformDevice(Device dev) {
		DeviceType ret = null;
		if (dev != null) {
			ret = new DeviceType();
			if (dev.getDeviceType() == DeviceTypeEnum.aikName) {
				ret.setAikName(dev.getValue());
			} else {
				ret.setName(dev.getValue());
			}
		}
		return ret;
	}

	
	/**
	 * From autogenerated Access Request to datamodel Access Request
	 * 
	 * @param art
	 * @return
	 * @throws IdentifierConstructionException
	 */
	AccessRequest transformAR(AccessRequestType art)
										throws IdentifierConstructionException {
		AccessRequest ret = null;
		if (art != null) {
			ret = new AccessRequest(art.getName(), art.getAdministrativeDomain());
		}
		return ret;
	}

	
	/**
	 * From datamodel Access Request to autogenerated Access Request
	 * 
	 * @param ar
	 * @return
	 */
	AccessRequestType transformAR(AccessRequest ar) {
		AccessRequestType ret = null;
		if (ar != null) { 
			ret = new AccessRequestType();
			ret.setName(ar.getName());
			if (ar.getAdministrativeDomain().length() == 0) {
				ret.setAdministrativeDomain(null);
			} else {
				ret.setAdministrativeDomain(ar.getAdministrativeDomain());
			}
		}
		return ret;
	}
}
