/*
 * Project: irond
 * Package: src.de.fhhannover.inform.iron.mapserver.utils
 * File:    JaxbIdentityTypeTransformer.java
 *
 * Copyright (C) 2010-2011 Fachhochschule Hannover
 * Ricklinger Stadtweg 118, 30459 Hannover, Germany 
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.fhhannover.inform.iron.mapserver.utils;

import de.fhhannover.inform.iron.mapserver.datamodel.exceptions.IdentifierConstructionException;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.IdentityTypeEnum;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.IpAddressTypeEnum;

/**
 * Simple class to transform from autogenerated type strings
 * to type enums. Type enums are used in the datamodel.
 * 
 * @author aw
 */
class JaxbIdentityTypeTransformer {
	
	private static JaxbIdentityTypeTransformer instance = new JaxbIdentityTypeTransformer();
	
	private JaxbIdentityTypeTransformer() { }
	
	static JaxbIdentityTypeTransformer getInstance() {
		if (instance == null)
			instance = new JaxbIdentityTypeTransformer();
		
		return instance;
	}
	
	
	/**
	 * Array of strings which represent the different contents
	 * of a identity type
	 */
	private final String identityTypeStrings[] = {	"aik-name",
													"distinguished-name",
													"dns-name",
													"email-address",
													"hip-hit",
													"kerberos-principal",
													"username",
													"sip-uri",
													"tel-uri",
													"other"  };
	
	/**
	 * Array of enums which the identity type string is mapped to
	 */
	private final IdentityTypeEnum identityTypeEnums[] = IdentityTypeEnum.values();

	
	/**
	 * Helper method to transform auto generated identity type strings to
	 * datamodel identity type enums
	 * 
	 * Depends on the right order of the enum and string array!
	 * 
	 * @param t
	 * @return
	 * @throws IdentifierConstructionException
	 */
	IdentityTypeEnum transformIdentityType(String type)
									throws  IdentifierConstructionException {
		if (type != null) {
			for (int i = 0; i < identityTypeStrings.length; i++) {
				if(identityTypeStrings[i].equals(type)) {
					return identityTypeEnums[i];
				}
			}
		}
		// if we come here something failed
		throw new IdentifierConstructionException("invalid identity type");
	}

	
	/**
	 * Helper method to transform datamodel identity type enums to
	 * autogenerated identity type strings
	 * 
	 * Depends on the right order of the enum and string array!
	 * 
	 * @param type
	 * @return
	 */
	String transformIdentityType(IdentityTypeEnum type) {
		if (type != null) {
			for (int i = 0; i < identityTypeEnums.length; i++) {
				if (identityTypeEnums[i] == type) {
					return new String(identityTypeStrings[i]);
				}
			}
		}
		return null;
	}
	

	/**
	 * Transform from a IpAddressType string into IpAdressType enum
	 * 
	 * @param ipt
	 * @return
	 */
	IpAddressTypeEnum transformIpAddressType(String ipt) {
		IpAddressTypeEnum ret = null;
		
		if (ipt != null) {
			if (ipt.equals(IpAddressTypeEnum.IPv4.toString())) {
				ret = IpAddressTypeEnum.IPv4;
			} else {
				ret = IpAddressTypeEnum.IPv6;
			}
		}
		return ret;
	}
	
	
	/**
	 * Convert a datamodel IpAddressType enum into an String
	 * containing the IpAddressType 
	 * 
	 * @param ipt
	 * @return
	 */
	String transformIpAddressType(IpAddressTypeEnum ipt) {
		String ret = null;
		if (ipt != null) {
			ret = new String(ipt.toString());
		}
		return ret;
	}
}
