/*
 * Project: irond
 * Package: src.de.fhhannover.inform.iron.mapserver.utils
 * File:    JaxbResponseCreator.java
 *
 * Copyright (C) 2010-2011 Fachhochschule Hannover
 * Ricklinger Stadtweg 118, 30459 Hannover, Germany 
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.fhhannover.inform.iron.mapserver.utils;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.InputStream;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.StringTokenizer;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;
import javax.xml.bind.PropertyException;

import org.trustedcomputinggroup._2010.ifmap._2.DeleteResultType;
import org.trustedcomputinggroup._2010.ifmap._2.DumpResponseType;
import org.trustedcomputinggroup._2010.ifmap._2.ErrorResultType;
import org.trustedcomputinggroup._2010.ifmap._2.MetadataListType;
import org.trustedcomputinggroup._2010.ifmap._2.NewSessionResultType;
import org.trustedcomputinggroup._2010.ifmap._2.NotifyResultType;
import org.trustedcomputinggroup._2010.ifmap._2.ObjectFactory;
import org.trustedcomputinggroup._2010.ifmap._2.PollResultType;
import org.trustedcomputinggroup._2010.ifmap._2.ResponseType;
import org.trustedcomputinggroup._2010.ifmap._2.ResponseType.PublishReceived;
import org.trustedcomputinggroup._2010.ifmap._2.ResponseType.PurgePublisherReceived;
import org.trustedcomputinggroup._2010.ifmap._2.ResponseType.RenewSessionResult;
import org.trustedcomputinggroup._2010.ifmap._2.ResponseType.SubscribeReceived;
import org.trustedcomputinggroup._2010.ifmap._2.ResultItemType;
import org.trustedcomputinggroup._2010.ifmap._2.SearchResultType;
import org.trustedcomputinggroup._2010.ifmap._2.UpdateResultType;
import org.trustedcomputinggroup._2010.ifmap._2.ResponseType.EndSessionResult;
import org.w3._2003._05.soap_envelope.Body;
import org.w3._2003._05.soap_envelope.Envelope;

import de.fhhannover.inform.iron.mapserver.datamodel.graph.AccessRequest;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.Device;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.Identifier;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.Identity;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.IpAddress;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.Link;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.MacAddress;
import de.fhhannover.inform.iron.mapserver.datamodel.graph.meta.Metadata;
import de.fhhannover.inform.iron.mapserver.messages.DumpResult;
import de.fhhannover.inform.iron.mapserver.messages.ErrorCode;
import de.fhhannover.inform.iron.mapserver.messages.PollResult;
import de.fhhannover.inform.iron.mapserver.messages.SearchResult;

/**
 * A implementation of {@link ResponseCreator} using JAXB as XML binding.
 * 
 * Originates from the old MessageTransformer class
 * 
 * @author aw
 */
class JaxbResponseCreator implements ResponseCreator {
	
	private JAXBContext mJaxbCtx; 
	private ThreadLocal<Marshaller> mMarshaller;

	private ObjectFactory mIfmapObjFac = new ObjectFactory();
	private org.w3._2003._05.soap_envelope.ObjectFactory mSoapObjFac =
			new org.w3._2003._05.soap_envelope.ObjectFactory();
	
	
	private JaxbIdentifierTransformer iTrans = JaxbIdentifierTransformer.getInstance();
	
	JaxbResponseCreator() {
		try {
			mJaxbCtx = JAXBContext.newInstance(Envelope.class);
						// shouldn't this be enough?
						//Header.class,
						//Body.class,
						//NewSessionRequestType.class);
			mMarshaller = new ThreadLocal<Marshaller>() {
				@Override
				protected Marshaller initialValue() {
					Marshaller ret =  null;
					try {
						ret =  mJaxbCtx.createMarshaller();
						ret.setProperty(Marshaller.JAXB_FORMATTED_OUTPUT, Boolean.TRUE);
					} catch (PropertyException e) {
						e.printStackTrace();
						// if it's not possible it's not possible
					} catch (JAXBException e) {
						throw new RuntimeException(e);
					}
					return ret;
				}
			};
		} catch (JAXBException e) {
			// go crazy if we can't initialize the JAXB context or something
			// JAXB related
			throw new RuntimeException(e);
		}
	}
	
	
	private InputStream doMarshal(Envelope env) {
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		try {
			// I have no idea why... Maybe a book on JAXB would be good...
			JAXBElement<Envelope> jaxbEnv = mSoapObjFac.createEnvelope(env);
			mMarshaller.get().marshal(jaxbEnv, baos);
		} catch (JAXBException e) {
			// why would this ever fail? if it does we have a problem anyway?
			e.printStackTrace();
			throw new RuntimeException(e.getMessage());
		}
		return new ByteArrayInputStream(baos.toByteArray());
	}
	
	private Envelope prepareEnvelopeWithResponse() {
		Envelope env = mSoapObjFac.createEnvelope();
		Body body = mSoapObjFac.createBody();
		env.setBody(body);
		ResponseType rt = mIfmapObjFac.createResponseType();
		body.setResponse(rt);
		return env;
	}
	
	
	private ResponseType getResponseElement(Envelope env) {
		return env.getBody().getResponse();
	}


	@Override
	public InputStream createNewSessionResponse(String sessionId, String publisherId) {
		Envelope env = prepareEnvelopeWithResponse();
		ResponseType resp = getResponseElement(env);
		NewSessionResultType nsrt = getNsrt(sessionId, publisherId);
		resp.setNewSessionResult(nsrt);
		return doMarshal(env);
	}

	@Override
	public InputStream createNewSessionResponse(String sessionId,
			String publisherId, long maxPollResultSize) {
		Envelope env = prepareEnvelopeWithResponse();
		ResponseType resp = getResponseElement(env);
		NewSessionResultType nsrt = getNsrt(sessionId, publisherId);
		nsrt.setMaxPollResultSize(new BigInteger(""+maxPollResultSize));
		resp.setNewSessionResult(nsrt);
		return doMarshal(env);
	}
	
	private NewSessionResultType getNsrt(String sessionId, String publisherId) {
		NewSessionResultType nsrt = mIfmapObjFac.createNewSessionResultType();
		nsrt.setSessionId(sessionId);
		nsrt.setIfmapPublisherId(publisherId);
		return nsrt;
	}
	
	
	@Override
	public InputStream createEndSessionResponse() {
		Envelope env = prepareEnvelopeWithResponse();
		ResponseType resp = getResponseElement(env);
		EndSessionResult esr = new EndSessionResult();
		resp.setEndSessionResult(esr);
		return doMarshal(env);
	}

	
	@Override
	public InputStream createRenewSessionResponse() {
		Envelope env = prepareEnvelopeWithResponse();
		ResponseType resp = getResponseElement(env);
		RenewSessionResult rsr = new RenewSessionResult();
		resp.setRenewSessionResult(rsr);
		return doMarshal(env);
	}

	@Override
	public InputStream createPublishResponse() {
		Envelope env = prepareEnvelopeWithResponse();
		ResponseType resp = getResponseElement(env);
		PublishReceived pr = new PublishReceived();
		resp.setPublishReceived(pr);
		return doMarshal(env);
	}

	@Override
	public InputStream createPurgePublisherResponse() {
		Envelope env = prepareEnvelopeWithResponse();
		ResponseType resp = getResponseElement(env);
		PurgePublisherReceived ppr = new PurgePublisherReceived();
		resp.setPurgePublisherReceived(ppr);
		return doMarshal(env);
	}

	
	@Override
	public InputStream createSubscribeResponse() {
		Envelope env = prepareEnvelopeWithResponse();
		ResponseType resp = getResponseElement(env);
		SubscribeReceived sr = new SubscribeReceived();
		resp.setSubscribeReceived(sr);
		return doMarshal(env);
	}

	
	@Override
	public InputStream createErrorResponse(ErrorCode errorCode, String errorString) {
		Envelope env = prepareEnvelopeWithResponse();
		ResponseType resp = getResponseElement(env);
		ErrorResultType ert = mIfmapObjFac.createErrorResultType();
		ert.setErrorCode(errorCode.toString());
		ert.setErrorString(errorString);
		resp.setErrorResult(ert);
		return doMarshal(env);
	}

	@Override
	public InputStream createSearchResponse(SearchResult searchResult) {
		Envelope env = prepareEnvelopeWithResponse();
		ResponseType resp = getResponseElement(env);
		SearchResultType srt = transformSearchResult(searchResult);
		resp.setSearchResult(srt);
		return doMarshal(env);
	}
	
	@Override
	public InputStream createDumpResponse(DumpResult dumpResult) {
		Envelope env = prepareEnvelopeWithResponse();
		ResponseType resp = getResponseElement(env);		
		DumpResponseType dump = transformDumpResult(dumpResult);
		resp.setDumpResult(dump);		
		return doMarshal(env);
	}

	@Override
	public InputStream createPollResponse(PollResult pollResult) {
		Envelope env = prepareEnvelopeWithResponse();
		ResponseType resp = getResponseElement(env);
		PollResultType prt = transformPollResult(pollResult);
		resp.setPollResult(prt);
		return doMarshal(env);
	}
	
	
	@Override
	public InputStream createPollResponseError(ErrorCode errCode, String errMsg,
			String subName) {
		Envelope env = prepareEnvelopeWithResponse();
		ResponseType resp = getResponseElement(env);
		PollResultType prt = mIfmapObjFac.createPollResultType();
		ErrorResultType ert = mIfmapObjFac.createErrorResultType();
		ert.setErrorCode(errCode.toString());
		ert.setErrorString(errMsg);
		ert.setName(subName);
		prt.getSearchResultOrUpdateResultOrDeleteResult().add(ert);
		resp.setPollResult(prt);
		return doMarshal(env);
	}

	
	private PollResultType transformPollResult (PollResult pr) {
		PollResultType prt = new PollResultType();

		// list with all results
		List<Object> list = prt.getSearchResultOrUpdateResultOrDeleteResult();
		// list with delete results
		List<SearchResult> deleteResults =  pr.getDeleteResult();
		List<SearchResult> updateResults =  pr.getUpdateResult();
		List<SearchResult> notifyResults =  pr.getNotifyResult();
		List<SearchResult> searchResults =  pr.getSearchResult();
		
		for (SearchResult sr : deleteResults) {
			list.add(transformDeleteResult(sr));
		}
		
		for (SearchResult sr : updateResults) {
			list.add(transformUpdateResult(sr));
		}
		
		for (SearchResult sr : notifyResults) {
			list.add(transformNotifyResult(sr));
		}

		for (SearchResult sr : searchResults) {
			list.add(transformSearchResult(sr));
		}
		
		
		return prt;
		
	}
	

	private DeleteResultType transformDeleteResult(SearchResult sr) {
		DeleteResultType res = null;
		if (sr != null) {
			res = new DeleteResultType();
			if (sr.getName() != null) {
				res.setName(sr.getName());
			}
			addResultItems(res, sr);
		}
		
		return res;
	}
	

	private UpdateResultType transformUpdateResult(SearchResult sr) {
		UpdateResultType res = null;
		if (sr != null) {
			res = new UpdateResultType();
			if (sr.getName() != null) {
				res.setName(sr.getName());
			}
			addResultItems(res, sr);
		}
		
		return res;
	}
	
	private NotifyResultType transformNotifyResult(SearchResult sr) {
		NotifyResultType res = null;
		if (sr != null) {
			res = new NotifyResultType();
			if (sr.getName() != null) {
				res.setName(sr.getName());
			}
			addResultItems(res, sr);
		}
		
		return res;
	}
	
	private SearchResultType transformSearchResult(SearchResult sr) {
		SearchResultType res = null;
		if (sr != null) {
			res = new SearchResultType();
			if (sr.getName() != null) {
				res.setName(sr.getName());
			}
			addResultItems(res, sr);
		}
		
		return res;
	}

	private DumpResponseType transformDumpResult(DumpResult dr) {		
		if (dr != null) {						
			String istr = dr.getFilter();				
	    	DumpResponseType dt =  new DumpResponseType();
	    	
	    	Identifier[] idents = dr.getIdentifier();
	    	long last_update = dr.getLastUpdateTime();
	    	
	    	dt.setLastUpdate(""+last_update);
	    	
	    	if(istr == null){
	    		istr = "";
	    	}
	    	else {
	    		istr = istr.trim().toLowerCase();
	    	}
	    		    	
			if(istr.equals("") ||  istr.equals("*")){
				for(Identifier id : idents){    			
	    			this.add(dt, id);
				}
			}
			else {		
				StringTokenizer tokenizer = new StringTokenizer(istr,",");
				ArrayList<String> list = new ArrayList<String>();
				while(tokenizer.hasMoreElements())
					list.add(tokenizer.nextToken());		    
				if(!istr.equals("-")){
					for(Identifier id : idents){
						//Skip subscriptions
						if(id.getLinks().size() > 0){								
							if(list.contains(id.getType().getTypeString()))
								this.add(dt, id);    			
						}
					}					
				}
			}
	    	
			return dt;
		}	
		return null;
	}
	
	private void add(DumpResponseType drt, Identifier identifier) {
		if (identifier != null && drt != null) {
			List<Object> retlist = drt.getAccessRequestOrIdentityOrIpAddress();
			if (identifier instanceof AccessRequest) {				
				retlist.add(iTrans.transformAR((AccessRequest)identifier));
			} else if (identifier instanceof Device) {
				retlist.add(iTrans.transformDevice((Device)identifier));
			} else if (identifier instanceof Identity) {
				retlist.add(iTrans.transformIdentity((Identity)identifier));
			} else if (identifier instanceof IpAddress) {				
				retlist.add(iTrans.transformIP((IpAddress)identifier));
			} else if (identifier instanceof MacAddress) {
				retlist.add(iTrans.transformMacAddress((MacAddress)identifier));
			}
		}
	}

	private void addResultItems(DeleteResultType res, SearchResult result) {
		if (result != null && res != null) {
			List<ResultItemType> rits = createResultItemList(result);
				res.getResultItem().clear();
				res.getResultItem().addAll(rits);
		}
	}


	private void addResultItems(UpdateResultType res, SearchResult result) {
		if (result != null && res != null) {
			List<ResultItemType> rits = createResultItemList(result);
				res.getResultItem().clear();
				res.getResultItem().addAll(rits);
		}
	}
	private void addResultItems(NotifyResultType res, SearchResult result) {
		if (result != null && res != null) {
			List<ResultItemType> rits = createResultItemList(result);
				res.getResultItem().clear();
				res.getResultItem().addAll(rits);
		}
	}
	
	private void addResultItems(SearchResultType res, SearchResult result) {
		if (result != null && res != null) {
			List<ResultItemType> rits = createResultItemList(result);
				res.getResultItem().clear();
				res.getResultItem().addAll(rits);
		}
	}
		
	private List<ResultItemType> createResultItemList(SearchResult result) {
		List<ResultItemType> list = new LinkedList<ResultItemType>();
		for (Identifier i : result.getIdentifiers())  {
			List<Metadata> dmlist = result.getMetadataFor(i);
			ResultItemType rit = transformToResultItem(i, dmlist);
			list.add(rit);
		}
			
		for (Link l : result.getLinks()) {
			List<Metadata> dmlist = result.getMetadataFor(l);
			ResultItemType rit = transformToResultItem(l, dmlist);
			list.add(rit);
		}
		return list;
	}
	

	/**
	 * Extract information from a link into a ResultItemType object,
	 * helper mehtod for transformSearchResult()
	 * 
	 * FIXME: Somehow redundant code with the method below
	 * 
	 * @param i
	 * @param items
	 * @param link
	 */
	private ResultItemType transformToResultItem(Link link, List<Metadata> dmlist) {
		ResultItemType rit = null;
		if (link != null) {
			rit = new ResultItemType();
			MetadataListType mlist;
			add(rit, link.getIdent1());
			add(rit, link.getIdent2());
			mlist = transformMetadata(dmlist);
			rit.setMetadata(mlist);
		}
		return rit;
	}

	
	/**
	 * Extract information from an identifier and put it into a
	 * ResultItemType object, helper mehtod for transformSearchResult()
	 * 
	 * FIXME: Somehow redundant code with the method above
	 * 
	 * @param item
	 * @param identifier
	 */
	private ResultItemType transformToResultItem(Identifier identifier, List<Metadata> dmlist) {
		ResultItemType rit = null;
		if (identifier != null) {
			rit = new ResultItemType();
			add(rit, identifier);
			MetadataListType mlist = transformMetadata(dmlist);
			rit.setMetadata(mlist);
		}
		return rit;
	}

	
	/**
	 * Transform a given identifier into an object of a autogenerated class
	 * and add it into the given ResultItemType object.
	 * 
	 * @param rit
	 * @param identifier
	 */
	private void add(ResultItemType rit, Identifier identifier) {
		if (identifier != null && rit != null) {
			List<Object> retlist = rit.getAccessRequestOrIdentityOrIpAddress();
			if (identifier instanceof AccessRequest) {
				retlist.add(iTrans.transformAR((AccessRequest)identifier));
			} else if (identifier instanceof Device) {
				retlist.add(iTrans.transformDevice((Device)identifier));
			} else if (identifier instanceof Identity) {
				retlist.add(iTrans.transformIdentity((Identity)identifier));
			} else if (identifier instanceof IpAddress) {
				retlist.add(iTrans.transformIP((IpAddress)identifier));
			} else if (identifier instanceof MacAddress) {
				retlist.add(iTrans.transformMacAddress((MacAddress)identifier));
			} else {
				throw new RuntimeException("Unknown Identifier class: " + 
						identifier.getClass().getCanonicalName());
			}
		}
	}
	
	/**
	 * This method converts from a datamodel metadata list to a
	 * MetadataListType.
	 * It simply uses the toW3CDocument() method of Metadata implementations
	 * and puts those into the MetaDataListType
	 * 
	 * Warning, returns null if no metadata is available in the 
	 * given list.
	 * 
	 * @param List of metadata from the datamodel
	 * @return Type which can be set in the autogenerated classes as
	 *         list
	 */
	MetadataListType transformMetadata(List<Metadata> metadata) {
		MetadataListType ret = null;
		if (metadata != null && metadata.size() > 0) {
			ret = new MetadataListType();
			
			for (Metadata m : metadata) {
				org.w3c.dom.Document w3cDoc =  m.toW3cDocument();
				if (w3cDoc != null) {
					ret.getAny().add(w3cDoc.getDocumentElement());
				}
			}
		}
		return ret;
	}
}
